"""Package-level data"""

import coloredlogs
import logmuse

from bedboss.bbuploader.constants import PKG_NAME

_LOGGER = logmuse.init_logger(PKG_NAME)
coloredlogs.install(
    logger=_LOGGER,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [BBUPLOADER] %(message)s",
)



import logging
import os
from typing import Dict, List, Optional, Union, Tuple

from gtars.models import RegionSet as GRegionSet

from bedboss.exceptions import ValidatorException, BedBossException

# from bedboss.refgenome_validator.const import GENOME_FILES
from bedboss.refgenome_validator.genome_model import GenomeModel
from bedboss.refgenome_validator.models import (
    ChromLengthStats,
    ChromNameStats,
    CompatibilityConcise,
    CompatibilityStats,
    RatingModel,
    SequenceFitStats,
)
from bedboss.refgenome_validator.utils import (
    get_bed_chrom_info,
    parse_IGD_output,
    run_igd_command,
    predict_from_compatibility_resutlts,
)
from bedboss.refgenome_validator.refgenie_chrom_sizes import get_chrom_sizes

_LOGGER = logging.getLogger("bedboss")


class ReferenceValidator:
    """
    Primary class for creating a compatibility dict
    """

    def __init__(
        self,
        genome_models: Optional[List[GenomeModel]] = None,
        igd_path: Optional[str] = None,
    ):
        """
        Initialization method

        :param genome_models: this is a list of GenomeModels that will be checked against a bed file. Default: None
        :param igd_path: path to a local IGD file containing ALL excluded ranges intervals for IGD overlap assessment,
            if not provided these metrics are not computed. Default: None
        """

        if not genome_models:
            genome_models = self._build_default_models()
        elif isinstance(genome_models, str):
            genome_models = list(genome_models)
        elif not isinstance(genome_models, list):
            raise ValidatorException(
                reason="A list of GenomeModels must be provided to initialize the Validator class"
            )

        self.genome_models: List[GenomeModel] = genome_models
        self.igd_path = igd_path

    @staticmethod
    def calculate_chrom_stats(
        bed_chrom_sizes: dict, genome_chrom_sizes: dict
    ) -> CompatibilityStats:
        """
        Calculate overlap and sequence fit.
        Stats are associated with comparison of chrom names, chrom lengths, and sequence fits.

        :param bed_chrom_sizes: dict of a bedfile's chrom size
        :param genome_chrom_sizes: dict of a GenomeModel's chrom sizes

        :return: dictionary with information on Query vs Model, e.g. chrom names QueryvsModel
        """

        # Layer 1: Check names and Determine XS (Extra Sequences) via Calculation of Recall/Sensitivity
        # Q = Query, M = Model
        q_and_m = 0  # how many chrom names are in the query and the genome model?
        q_and_not_m = (
            0  # how many chrom names are in the query but not in the genome model?
        )
        not_q_and_m = (
            0  # how many chrom names are not in the query but are in the genome model?
        )
        passed_chrom_names = True  # does this bed file pass the first layer of testing?

        query_keys_present = []  # These keys are used for seq fit calculation
        for key in list(bed_chrom_sizes.keys()):
            if key not in genome_chrom_sizes:
                q_and_not_m += 1
            if key in genome_chrom_sizes:
                q_and_m += 1
                query_keys_present.append(key)

        for key in list(genome_chrom_sizes.keys()):
            if key not in bed_chrom_sizes:
                not_q_and_m += 1

        # Calculate the Jaccard Index for Chrom Names
        bed_chrom_set = set(list(bed_chrom_sizes.keys()))
        genome_chrom_set = set(list(genome_chrom_sizes.keys()))
        chrom_intersection = bed_chrom_set.intersection(genome_chrom_set)
        chrom_union = bed_chrom_set.union(chrom_intersection)
        chrom_jaccard_index = len(chrom_intersection) / len(chrom_union)

        # Alternative Method for Calculating Jaccard_index for binary classification
        # JI = TP/(TP+FP+FN)
        jaccard_binary = q_and_m / (q_and_m + not_q_and_m + q_and_not_m)

        # What is our threshold for passing layer 1?
        if q_and_not_m > 0:
            passed_chrom_names = False

        # Calculate sensitivity for chrom names
        # XS -> Extra Sequences
        sensitivity = q_and_m / (q_and_m + q_and_not_m)

        name_stats = ChromNameStats(
            xs=sensitivity,
            q_and_m=q_and_m,
            q_and_not_m=q_and_not_m,
            not_q_and_m=not_q_and_m,
            jaccard_index=chrom_jaccard_index,
            jaccard_index_binary=jaccard_binary,
            passed_chrom_names=passed_chrom_names,
        )

        # Layer 2:  Check Lengths, but only if layer 1 is passing [all chroms are in ref genome]
        if passed_chrom_names:
            chroms_beyond_range = False
            num_of_chrom_beyond = 0
            num_chrom_within_bounds = 0

            for key in list(bed_chrom_sizes.keys()):
                if key in genome_chrom_sizes:
                    if bed_chrom_sizes[key] > genome_chrom_sizes[key]:
                        num_of_chrom_beyond += 1
                        chroms_beyond_range = True
                    else:
                        num_chrom_within_bounds += 1

            # Calculate recall/sensitivity for chrom lengths
            # OOBR -> Out of Bounds Range
            sensitivity = num_chrom_within_bounds / (
                num_chrom_within_bounds + num_of_chrom_beyond
            )
            length_stats = ChromLengthStats(
                oobr=sensitivity,
                beyond_range=chroms_beyond_range,
                num_of_chrom_beyond=num_of_chrom_beyond,
                percentage_bed_chrom_beyond=(
                    100 * num_of_chrom_beyond / len(bed_chrom_set)
                ),
                percentage_genome_chrom_beyond=(
                    100 * num_of_chrom_beyond / len(genome_chrom_set)
                ),
            )

        else:
            length_stats = ChromLengthStats()

        # Layer 3 Calculate Sequence Fit if any query chrom names were present
        if len(query_keys_present) > 0:
            bed_sum = 0
            ref_genome_sum = 0
            for q_chr in query_keys_present:
                bed_sum += int(genome_chrom_sizes[q_chr])
            for g_chr in genome_chrom_sizes:
                ref_genome_sum += int(genome_chrom_sizes[g_chr])

            seq_fit_stats = SequenceFitStats(sequence_fit=bed_sum / ref_genome_sum)

        else:
            seq_fit_stats = SequenceFitStats(sequence_fit=None)

        return CompatibilityStats(
            chrom_name_stats=name_stats,
            chrom_length_stats=length_stats,
            chrom_sequence_fit_stats=seq_fit_stats,
        )

    def get_igd_overlaps(self, bedfile: str) -> Union[dict[str, dict], dict[str, None]]:
        """
        Third layer compatibility check.
        Run helper functions and execute an igd search query across an Integrated Genome Database.

        :param bedfile: path to the bedfile
        :return: dict of dicts containing keys (file names) and values (number of overlaps). Or if no overlaps are found,
        it returns an empty dict.

        # TODO: should be a pydantic model

        Currently for this function to work, the user must install the C version of IGD and have created a local igd file
        for the Excluded Ranges Bedset:
        https://github.com/databio/IGD
        https://bedbase.org/bedset/excluderanges

        """
        if not self.igd_path:
            return {"igd_stats": None}

        try:
            IGD_LOCATION = os.environ["IGD_LOCATION"]
        except:
            # Local installation of C version of IGD
            IGD_LOCATION = "/home/drc/GITHUB/igd/IGD/bin/igd"

        # Construct an IGD command to run as subprocess
        igd_command = IGD_LOCATION + f" search {self.igd_path} -q {bedfile}"

        returned_stdout = run_igd_command(igd_command)

        if not returned_stdout:
            return {"igd_stats": None}

        igd_overlap_data = parse_IGD_output(returned_stdout)

        if not igd_overlap_data:
            return {
                "igd_stats": {}
            }  # None tells us if the bed file never made it to layer 3 or perhaps igd errord, empty dict tells us that there were no overlaps found
        else:
            overlaps_dict = {}
            for datum in igd_overlap_data:
                if "file_name" in datum and "number_of_hits" in datum:
                    overlaps_dict.update({datum["file_name"]: datum["number_of_hits"]})

        return overlaps_dict

    def determine_compatibility(
        self,
        bedfile: Union[GRegionSet, str, Dict[str, int]],
        ref_filter: Optional[List[str]] = None,
        concise: Optional[bool] = False,
    ) -> Union[Dict[str, CompatibilityStats], Dict[str, CompatibilityConcise]]:
        """
        Determine compatibility of the bed file.

        :param bedfile: path to bedfile
        :param ref_filter: list of ref genome aliases to filter on.
        :param concise: if True, only return a concise list of compatibility stats. Default: False
        :return: a dict with CompatibilityStats, or CompatibilityConcise model (depends if concise is set to True)
        """

        if isinstance(bedfile, GRegionSet):
            _LOGGER.info(f"Calculating reference genome stats for {bedfile}...")
        elif isinstance(bedfile, str):
            _LOGGER.info(f"Calculating reference genome stats for {bedfile}...")
            bedfile = GRegionSet(bedfile)
        else:
            _LOGGER.info(
                f"Calculating reference genome stats for provided bed chrom dict..."
            )

        if ref_filter:
            # Filter out unwanted reference genomes to assess
            for genome_model in self.genome_models:
                if genome_model.genome_alias in ref_filter:
                    self.genome_models.remove(
                        genome_model
                    )  # TODO: remove it only for this analysis, not permanently
        try:
            if isinstance(bedfile, dict):
                bed_chrom_info = bedfile
            else:
                bed_chrom_info = get_bed_chrom_info(bedfile)
        except Exception as e:
            raise BedBossException(
                f"Unable to open bed file or determine compatibility. Error: {str(e)}"
            )

        if not bed_chrom_info:
            raise ValidatorException("Incorrect bed file provided")

        model_compat_stats = {}

        for genome_model in self.genome_models:
            # First and Second Layer of Compatibility
            model_compat_stats[genome_model.genome_digest]: CompatibilityStats = (
                self.calculate_chrom_stats(bed_chrom_info, genome_model.chrom_sizes)
            )

            # Third layer - IGD, only if layer 1 and layer 2 have passed
            if (
                model_compat_stats[
                    genome_model.genome_digest
                ].chrom_name_stats.passed_chrom_names
                and not model_compat_stats[
                    genome_model.genome_digest
                ].chrom_length_stats.beyond_range
            ):
                model_compat_stats[genome_model.genome_digest].igd_stats = (
                    self.get_igd_overlaps(bedfile)
                )

            # Calculate compatibility rating
            model_compat_stats[genome_model.genome_digest].compatibility = (
                self.calculate_rating(model_compat_stats[genome_model.genome_digest])
            )
        if concise:
            concise_dict = {}
            for name, stats in model_compat_stats.items():
                concise_dict[name] = self._create_concise_output(stats)

            return concise_dict

        return model_compat_stats

    def calculate_rating(self, compat_stats: CompatibilityStats) -> RatingModel:
        """
        Determine the compatibility tier

        Tiers:
            - Tier1: Excellent compatibility, 0 pts
            - Tier2: Good compatibility, may need some processing, 1-3 pts
            - Tier3: Bed file needs processing to work (shifted hg38 to hg19?), 4-6 pts
            - Tier4: Poor compatibility, 7-9 pts

        :param compat_stats: CompatibilityStats with unprocess compatibility statistics
        :return: RatingModel - {
                    assigned_points: int
                    tier_ranking: int
                }
        """

        points_rating = 0

        # 1. Check extra sequences sensitivity.
        # sensitivity = 1 is considered great and no points should be assigned
        xs = compat_stats.chrom_name_stats.xs
        if xs < 0.3:
            points_rating += 6  # 3 + 1 + 1 + 1
        elif xs < 0.5:
            points_rating += 5  # 3 + 1 + 1
        elif xs < 0.7:
            points_rating += 4  # 3 + 1
        elif xs < 1:
            points_rating += 3
        else:
            pass

        # 2. Check OOBR and assign points based on sensitivity
        # only assessed if no extra chroms in query bed file
        if compat_stats.chrom_name_stats.passed_chrom_names:
            oobr = compat_stats.chrom_length_stats.oobr

            if oobr < 0.3:
                points_rating += 6  # 3 + 1 + 1 + 1
            elif oobr < 0.5:
                points_rating += 5  # 3 + 1 + 1
            elif oobr < 0.7:
                points_rating += 4  # 3 + 1
            elif oobr < 1:
                points_rating += 3
        else:
            # Do nothing here, points have already been added when Assessing XS if it is not == 1
            pass

        # 3. Check Sequence Fit - comparing lengths in queries vs lengths of queries in ref genome vs not in ref genome
        sequence_fit = compat_stats.chrom_sequence_fit_stats.sequence_fit
        if sequence_fit:
            # since this is only on keys present in both, ratio should always be less than 1
            # Should files be penalized here or actually awarded but only if the fit is really good?
            if sequence_fit < 0.90:
                points_rating += 1
            if sequence_fit < 0.60:
                points_rating += 1
            if sequence_fit < 0.60:
                points_rating += 1

        else:
            # if no chrom names were found during assessment
            points_rating += 4

        # Run analysis on igd_stats
        # WIP, currently only showing IGD stats for informational purposes
        if compat_stats.igd_stats and compat_stats.igd_stats != {}:
            self._process_igd_stats(compat_stats.igd_stats)

        tier_ranking = 0
        if points_rating == 0:
            tier_ranking = 1
        elif 1 <= points_rating <= 3:
            tier_ranking = 2
        elif 4 <= points_rating <= 6:
            tier_ranking = 3
        elif 7 <= points_rating:
            tier_ranking = 4
        else:
            _LOGGER.info(
                f"Catching points discrepancy,points = {points_rating}, assigning to Tier 4"
            )
            tier_ranking = 4

        return RatingModel(assigned_points=points_rating, tier_ranking=tier_ranking)

    def _process_igd_stats(self, igd_stats: dict):
        """
        Placeholder to process IGD Stats and determine if it should impact tier rating
        """
        ...

    @staticmethod
    def _build_default_models() -> list[GenomeModel]:
        """
        Build a default list of GenomeModels from the chrom.sizes folder.
        Uses file names as genome alias.

        return list[GenomeModel]
        """

        # OLD: TO BE DELETED IN FUTURE RELEASES
        # dir_path = os.path.dirname(os.path.realpath(__file__))
        #
        # chrm_sizes_directory = os.path.join(dir_path, "chrom_sizes")
        #
        # all_genome_models = []
        # for root, dirs, files in os.walk(chrm_sizes_directory):
        #     for file in files:
        #         if file.endswith(".sizes"):
        #             curr_genome_model = GenomeModel(
        #                 genome_alias=file, chrom_sizes_file=os.path.join(root, file)
        #             )
        #             all_genome_models.append(curr_genome_model)
        #
        # return all_genome_models
        return get_chrom_sizes()

    @staticmethod
    def _create_concise_output(output: CompatibilityStats) -> CompatibilityConcise:
        """
        Convert extended CompatibilityStats to concise output

        :param output: full compatibility stats
        :return:  concise compatibility stats
        """
        return CompatibilityConcise(
            xs=output.chrom_name_stats.xs,
            oobr=output.chrom_length_stats.oobr,
            sequence_fit=output.chrom_sequence_fit_stats.sequence_fit,
            assigned_points=output.compatibility.assigned_points,
            tier_ranking=output.compatibility.tier_ranking,
        )

    def predict(
        self, bedfile: Union[str, GRegionSet]
    ) -> Tuple[Union[str, None], Union[str, None]]:
        """
        Predict compatibility of a bed file with reference genomes

        :param bedfile: path to bedfile

        :return: tuple of (genome_id, digest) for the best compatible reference genome,
            or None if no tier 1 compatibility is found or digest doesn't exist
        """

        _LOGGER.info(f"Predicting compatibility of {bedfile} with reference genomes...")
        compatibility_stats: Dict[str, CompatibilityConcise] = (
            self.determine_compatibility(bedfile, concise=True)
        )

        return predict_from_compatibility_resutlts(compatibility_stats)



import sys

import logging

from bedboss.cli import app
from bedboss.const import PKG_NAME

_LOGGER = logging.getLogger(PKG_NAME)


def main():
    app(prog_name=PKG_NAME)


if __name__ == "__main__":
    try:
        main()

    except KeyboardInterrupt:
        print("Pipeline aborted.")
        sys.exit(1)



from qdrant_client import QdrantClient
import os
import sys
import pandas as pd
import logging
import numpy as np
from pydantic import BaseModel, ConfigDict

from umap import UMAP
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

import matplotlib.pyplot as plt
import seaborn as sns
from typing import Union
import warnings
from functools import lru_cache

import joblib
from bbconf import BedBaseAgent
import json

from bedboss.const import PKG_NAME

_LOGGER = logging.getLogger(PKG_NAME)

python_version = f"{sys.version_info.major}_{sys.version_info.minor}"


class umapReturn(BaseModel):
    model: Union[UMAP, PCA, TSNE]
    dataframe: pd.DataFrame

    model_config = ConfigDict(arbitrary_types_allowed=True)


def save_umap_model(umap_model: Union[UMAP, PCA, TSNE], model_path: str) -> None:
    """
    Save the UMAP, PCA, or t-SNE model to a file.

    :param umap_model: Fitted UMAP, PCA, or t-SNE model
    :param model_path: Path to save the model
    :return: None

    """
    with open(model_path, "wb") as file:
        joblib.dump(umap_model, file)

    _LOGGER.info(f"Model saved to {model_path}")


# @lru_cache()
# TODO: can we make this function cached, without using credentials as part of the cache key?
def fetch_data(agent: BedBaseAgent) -> pd.DataFrame:
    """
    Fetch data from Qdrant collection and return it as a DataFrame.

    :param agent: BedBaseAgent instance containing Qdrant access details

    :return: DataFrame with the following columns:
    """

    _LOGGER.info("Fetching data from Qdrant...")

    client = QdrantClient(
        host=agent.config.config.qdrant.host,
        port=6333,
        api_key=agent.config.config.qdrant.api_key,
    )

    points = []
    next_offset = 0
    batch_size = 10000

    while next_offset is not None:
        response = client.scroll(
            collection_name=agent.config.config.qdrant.file_collection,
            limit=batch_size,
            offset=next_offset,
            with_payload=True,
            with_vectors=True,
        )
        batch_points, next_offset = response
        points.extend(batch_points)
        _LOGGER.info(
            f"Fetched batch of {len(batch_points)} records (total: {len(points)})."
        )

    for m in points:
        m.id = m.id.replace("-", "")
    payload_df = pd.DataFrame([{**m.payload, "vector": m.vector} for m in points])
    merged = payload_df.set_index("id")

    _LOGGER.info(f"Fetched {len(merged)} records from Qdrant.")
    return merged


def save_df_as_json(df: pd.DataFrame, output_path: str) -> None:
    """
    Save a DataFrame as a JSON file in the specified format.
    It includes the following columns:
    - x, y, z (coordinates)
    - id (string identifier)
    - name (string)
    - description (string)
    - assay (string)
    - cell_line (string)

    :param df: DataFrame to save
    :param output_path: Path to save the JSON file
    :return: None

    """
    # Select the required columns
    columns_to_include = [
        "x",
        "y",
        "id",
        "name",
        "description",
        # "data_format",
        # "bed_compliance",
        "assay",
        "cell_line",
    ]

    if "z" in df.columns:
        columns_to_include.insert(2, "z")

    output_path = os.path.abspath(output_path)
    (f"Saving DataFrame as JSON to {output_path}")

    df.loc[:, "id"] = df.index.astype(str)  # Ensure 'id' is a string
    df = df.fillna("")

    coord_cols = [c for c in ["x", "y", "z"] if c in df.columns]

    # Create the nodes structure
    nodes = df[columns_to_include].to_dict(orient="records")
    for node in nodes:
        for col in coord_cols:
            if col in node:
                node[col] = round(float(node[col]), 2)

    # Create the final JSON structure
    json_data = {"nodes": nodes, "links": []}

    # Save to a JSON file
    output_path = f"{output_path}_{python_version}.json"
    with open(output_path, "w") as json_file:
        json.dump(json_data, json_file, indent=4)

    _LOGGER.info(f"Data saved to {output_path} successfully.")


def create_umap(
    df: pd.DataFrame,
    n_components: int = 3,
    plot_name: Union[str, None] = None,
    label_column: str = "cell_line",
    method: str = "umap",
) -> umapReturn:
    """
    Create UMAP, PCA, or t-SNE embeddings from the DataFrame.

    :param df: DataFrame containing the vectors and labels
    :param n_components: Number of dimensions for UMAP/PCA/t-SNE (default is 3)
    :param plot_name: Name for the output plot file. if None, no plot will be saved
    :param label_column: Column name to use for labeling the points in the plot e.g. "cell_line" or "assay". Default is "cell_line"
    :param method: Dimensionality reduction method to use. Options: "umap" (default), "pca", or "tsne"

    :return: Tuple of DataFrame with added coordinates and fitted model
    """

    if n_components not in [2, 3]:
        raise ValueError("n_components must be either 2 or 3.")

    method = method.lower()
    if method not in ["umap", "pca", "tsne"]:
        raise ValueError("method must be either 'umap', 'pca', or 'tsne'.")

    _LOGGER.info(f"Creating {method.upper()} embeddings...")

    if method == "umap":
        model = UMAP(
            n_components=n_components,
            random_state=42,
            verbose=True,
        )
    elif method == "pca":
        model = PCA(
            n_components=n_components,
            random_state=42,
        )
    else:  # method == "tsne"
        model = TSNE(
            n_components=n_components,
            random_state=42,
            verbose=True,
        )

    # Convert list of vectors to numpy array
    vectors = np.array(list(df["vector"]))

    if method == "tsne":
        # t-SNE doesn't support separate fit/transform, use fit_transform
        embeddings = model.fit_transform(vectors)
        fitted_model = model
    else:
        fitted_model = model.fit(vectors)
        embeddings = model.transform(vectors)

    if plot_name:
        if n_components == 2:
            if label_column not in ["cell_line", "assay"]:
                raise ValueError(
                    f"label_column must be either 'cell_line' or 'assay', got {label_column}."
                )

            plot_umap(embeddings, list(df[label_column]), name=plot_name)
        else:
            warnings.warn(
                "Plotting is only supported for 2D embeddings. No plot will be saved."
            )

    if n_components == 2:
        df[["x", "y"]] = pd.DataFrame(embeddings, index=df.index)
    elif n_components == 3:
        df[["x", "y", "z"]] = pd.DataFrame(embeddings, index=df.index)

    _LOGGER.info(f"{method.upper()} shape: {embeddings.shape}")

    return umapReturn(
        model=fitted_model,
        dataframe=df,
    )


def plot_umap(value, label, name="default") -> None:
    """
    Plot UMAP embeddings and save the figure.

    :param value: UMAP embeddings (only 2D)
    :param label: Labels for the points in the UMAP plot
    :param name: Name for the output plot file
    :return: None
    """

    fig, ax = plt.subplots(figsize=(5, 5))
    sns.scatterplot(
        ax=ax,
        x=value[:, 0],
        y=value[:, 1],
        hue=label,
        palette="tab20",
        s=2,  # Increase the size of the points
        edgecolor="none",
    )
    ax.legend(
        title=f"Clustering by {name}",
        loc="upper left",
        bbox_to_anchor=(1, 1),
        markerscale=2,
        fontsize="small",
    )
    ax.margins(0)  # Remove margins from the plot

    save_path = "."
    fig.savefig(
        os.path.join(save_path, f"{name}"),
        dpi=3000,
        bbox_inches="tight",
    )


def get_embeddings(
    bbconf: str,
    output_file: str,
    n_components: int = 2,
    plot_name: str = None,
    plot_label: str = None,
    top_assays: Union[int, None] = 15,
    top_cell_lines: Union[int, None] = 15,
    save_model: bool = True,
    method: str = "umap",
) -> None:
    """
    Get embeddings from Qdrant and create UMAP, PCA, or t-SNE, and save the results to a JSON file, and optionally plot the embeddings.

    :param bbconf: string containing to bedbase configuration file path
    :param output_file: Path to save the output JSON file
    :param n_components: Number of dimensions for UMAP/PCA/t-SNE [default is 2]

    :param plot_name: Name for the output plot file. if None, no plot will be saved
    :param plot_label: Column name to use for labeling the points in the plot e.g. "cell_line" or "assay". [Default is "cell_line"]

    :param top_assays: Number of top assays to consider. If None, all assays are considered. [Default is 15]
    :param top_cell_lines: Number of top cell lines to consider. If None, all. [Default is 15]

    :param save_model: Whether to save the model or not [Default is True]
    :param method: Dimensionality reduction method to use. Options: "umap", "pca", or "tsne" [Default is "umap"]

    :return: None

    """

    if isinstance(bbconf, str):
        agent = BedBaseAgent(config=bbconf)
    elif isinstance(bbconf, BedBaseAgent):
        agent = bbconf
    else:
        raise TypeError(
            "bbconf must be either a string path or a BedBaseAgent instance."
        )

    merged = fetch_data(agent=agent)

    if output_file.endswith(".json"):
        output_file = output_file[:-5]
        # output_file += ".json"
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    CELL_LINE = "cell_line"
    ASSAY = "assay"

    return_df = merged.copy()

    ##############################################################
    ######## Option 1 ############################################
    ######### Remove empty/None cell lines and assays ############
    ##############################################################

    # # Select top cell lines available in the dataset
    # if top_cell_lines is not None:
    #     top_cell_lines_list = [
    #         x
    #         for x in merged[CELL_LINE].value_counts().nlargest(top_cell_lines).index
    #         if x is not None and x != ""
    #     ]
    #
    #     return_df = return_df[return_df[CELL_LINE].isin(top_cell_lines_list)]
    #
    # # Select top assays available in the dataset
    # if top_assays is not None:
    #     top_assays_list = [
    #         x
    #         for x in merged[ASSAY].value_counts().nlargest(top_assays).index
    #         if x is not None and x != ""
    #     ]
    #
    #     return_df = return_df[return_df[ASSAY].isin(top_assays_list)]

    ################################################################################
    ######################### Option 2 #############################################
    ## Label empty/None cell lines and assays as "na" instead of removing them #####
    ################################################################################
    na_name = "UNKNOWN"

    return_df[CELL_LINE] = return_df[CELL_LINE].fillna(na_name).replace("", na_name)
    return_df[ASSAY] = return_df[ASSAY].fillna(na_name).replace("", na_name)

    # Select top cell lines available in the dataset
    if top_cell_lines is not None:
        top_cell_lines_list = list(
            return_df[CELL_LINE].value_counts().nlargest(top_cell_lines).index
        )

        return_df = return_df[return_df[CELL_LINE].isin(top_cell_lines_list)]

    # Select top assays available in the dataset
    if top_assays is not None:
        top_assays_list = list(
            return_df[ASSAY].value_counts().nlargest(top_assays).index
        )
        return_df = return_df[return_df[ASSAY].isin(top_assays_list)]

    ###############################################################################

    umap_return = create_umap(
        return_df,
        n_components=n_components,
        plot_name=plot_name,
        label_column=plot_label,
        method=method,
    )
    save_df_as_json(umap_return.dataframe, output_file)

    if save_model:
        ## controls the random initialization and stochastic optimization during UMAP fitting. But removing, because it causes issues during saving/loading
        ## I am removing it here, but it should be set later to the same value (42)
        if method == "umap":
            umap_return.model.random_state = None
        save_umap_model(
            umap_return.model,
            model_path=f"{output_file}_{method}_model_{python_version}.joblib",
        )

    _LOGGER.info(f"{method.upper()} processing completed!")



import json
import logging
import os
import subprocess
from typing import List, Union

import pephubclient
import peppy
from bbconf import BedBaseAgent
from bbconf.models.base_models import FileModel
from bbconf.models.bedset_models import BedSetPlots
from geniml.bbclient import BBClient
from geniml.io import BedSet
from pephubclient.helpers import is_registry_path

from bedboss.exceptions import BedBossException

_LOGGER = logging.getLogger("bedboss")


def create_bed_list_file(bedset: BedSet, file_path: str) -> None:
    """
    Create a file with bed_set_list (Later this file is used in R script)

    :param bedset: bed_set object
    :param file_path: path to the file
    :return: None
    """
    list_of_samples = [sample.path for sample in bedset]

    with open(file_path, "w") as f:
        for sample in list_of_samples:
            f.write(sample + "\n")

    return None


def create_plots(
    bedset: List[str],
    output_folder: str,
) -> dict:
    """
    Create plots for a bedset (commonality region plot)

    :param bedset: list of bedfiles ids
    :param output_folder: path to the output folder
    :return: dict with information about crated plots
    """
    bbclient_obj = BBClient()

    bed_set_object = BedSet()
    for bed_id in bedset:
        bed_set_object.add(bbclient_obj.load_bed(bed_id))

    bedset_md5sum = bed_set_object.identifier

    # if output folder doesn't exist create it
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    bedset_list_path = os.path.join(output_folder, f"{bedset_md5sum}_bedset.txt")
    create_bed_list_file(bed_set_object, bedset_list_path)
    rscript_path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        "bedbuncher",
        "tools",
        "bedsetStat.R",
    )
    assert os.path.exists(rscript_path), FileNotFoundError(
        f"'{rscript_path}' script not found"
    )

    json_file_path = os.path.join(output_folder, bedset_md5sum + ".json")
    command = (
        f"Rscript {rscript_path} --outputfolder={output_folder} "
        f"--bedfilelist={bedset_list_path} --id={bedset_md5sum} "
        f"--json={json_file_path}"
    )

    subprocess.run(command, shell=True)

    with open(json_file_path, "r", encoding="utf-8") as f:
        bedset_summary_info = json.loads(f.read())

    os.remove(bedset_list_path)
    os.remove(json_file_path)

    _LOGGER.info("Plots were created successfully and mediated files were removed")
    return bedset_summary_info["plots"][0]


def run_bedbuncher(
    bedbase_config: Union[str, BedBaseAgent],
    record_id: str,
    bed_set: List[str],
    output_folder: str,
    name: str = None,
    description: str = None,
    annotation: dict = None,
    heavy: bool = False,
    upload_pephub: bool = False,
    upload_s3: bool = False,
    no_fail: bool = False,
    force_overwrite: bool = False,
    lite: bool = False,
) -> None:
    """
    Add bedset to the database

    :param bedbase_config: BedBaseConf object
    :param record_id: record identifier or name to be used in database
    :param name: name of the bedset
    :param output_folder: path to the output folder
    :param bed_set: Bedset object or list of bedfiles ids
    :param description: Bedset description
    :param annotation: Bedset annotation (author, source, summary, etc.)
    :param heavy: whether to use heavy processing (add all columns to the database).
        if False -> R-script won't be executed, only basic statistics will be calculated
    :param no_fail: whether to raise an error if bedset was not added to the database
    :param upload_pephub: whether to create a view in pephub
    :param upload_s3: whether to upload files to s3
    :param force_overwrite: whether to overwrite the record in the database
    :param lite: whether to run the pipeline in lite mode
    # TODO: force_overwrite is not working!!! Fix it!
    :return:
    """
    _LOGGER.info(f"Adding bedset { record_id} to the database")

    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(bedbase_config)
    else:
        bbagent = bedbase_config

    if not record_id:
        raise BedBossException(
            "bedset_name was not provided correctly. Please provide it in pep name or as argument"
        )

    output_folder = os.path.join(
        output_folder,
        "bedsets",
    )

    if heavy:
        _LOGGER.info("Heavy processing is True. Calculating plots...")
        plot_value = create_plots(
            bedset=bed_set,
            output_folder=output_folder,
        )
        plots = BedSetPlots(region_commonality=FileModel(**plot_value))
    else:
        _LOGGER.info("Heavy processing is False. Plots won't be calculated")
        plots = None

    bbagent.bedset.create(
        identifier=record_id,
        name=name,
        bedid_list=bed_set,
        statistics=True,
        description=description,
        upload_pephub=upload_pephub,
        upload_s3=upload_s3,
        plots=plots.model_dump(exclude_none=True, exclude_unset=True) if plots else {},
        local_path=output_folder,
        no_fail=no_fail,
        overwrite=force_overwrite,
        annotation=annotation,
        processed=not lite,
    )


def run_bedbuncher_form_pep(
    bedbase_config: str,
    bedset_pep: Union[str, peppy.Project],
    output_folder: str,
    bedset_name: str = None,
    heavy: bool = False,
    upload_pephub: bool = False,
    upload_s3: bool = False,
    no_fail: bool = False,
    force_overwrite: bool = False,
) -> str:
    """
    Create bedset from pep and add it to the database

    :param bedbase_config: BedBaseConf object or path to the config file
    :param bedset_pep: path to the pep file or pephub registry path
    :param bedset_name: name of the bedset
    :param output_folder: path to the output folder
    :param heavy: whether to use heavy processing (add all columns to the database).
        if False -> R-script won't be executed, only basic statistics will be calculated
    :param upload_pephub: whether to create a view in pephub
    :param upload_s3: whether to upload files to s3
    :param no_fail: whether to raise an error if bedset was not added to the database
    :param force_overwrite: whether to overwrite the record in the database

    return bedset_name
    """
    if isinstance(bedset_pep, peppy.Project):
        pep_of_bed = bedset_pep
    elif isinstance(bedset_pep, str):
        if is_registry_path(bedset_pep):
            pep_of_bed = pephubclient.PEPHubClient().load_project(bedset_pep)
        else:
            pep_of_bed = peppy.Project(bedset_pep)
    else:
        raise ValueError(
            "bedset_pep should be either path to the pep file or pephub registry path"
        )

    bedfiles_list = [
        bedfile_id.get("record_identifier") or bedfile_id.sample_name
        for bedfile_id in pep_of_bed.samples
    ]

    run_bedbuncher(
        bedbase_config=bedbase_config,
        record_id=bedset_name or pep_of_bed.name,
        bed_set=bedfiles_list,
        output_folder=output_folder,
        name=bedset_name or pep_of_bed.name,
        description=pep_of_bed.description,
        heavy=heavy,
        upload_pephub=upload_pephub,
        upload_s3=upload_s3,
        no_fail=no_fail,
        force_overwrite=force_overwrite,
    )
    return bedset_name


# if __name__ == "__main__":
#     run_bedbuncher(
#         "/home/bnt4me/virginia/repos/bbuploader/config_db_local.yaml",
#         record_id="test_bedset",
#         name="This is a name",
#         bed_set=[
#             "bbad85f21962bb8d972444f7f9a3a932",
#             "0dcdf8986a72a3d85805bbc9493a1302",
#         ],
#         output_folder="/home/bnt4me/virginia/",
#         description="This is a description",
#         upload_s3=True,
#         no_fail=True,
#         upload_pephub=True,
#         heavy=True,
#     )

# run_bedbuncher_form_pep(
#     "/home/bnt4me/virginia/repos/bbuploader/config_db_local.yaml",
#     bedset_name="pephub_test",
#     bedset_pep="khoroshevskyi/bedbunch:default",
#     upload_pephub=True,
# )



import gzip
import logging
import os
import shutil
from pathlib import Path
from typing import Union, Tuple

import pypiper
from geniml.bbclient import BBClient
from gtars.models import RegionSet
from refgenconf.exceptions import MissingGenomeError
from ubiquerg import is_command_callable

from bedboss.bedclassifier.bedclassifier import get_bed_classification
from bedboss.bedmaker.const import (
    BEDGRAPH_TEMPLATE,
    BIGBED_FOLDER_NAME,
    BIGBED_TEMPLATE,
    BIGBED_TO_BED_PROGRAM,
    BIGWIG_TEMPLATE,
    WIG_TEMPLATE,
)
from bedboss.const import MAX_FILE_SIZE, MAX_REGION_NUMBER, MIN_REGION_WIDTH
from bedboss.bedmaker.models import BedMakerOutput, InputTypes
from bedboss.bedmaker.utils import get_chrom_sizes
from bedboss.exceptions import BedBossException, RequirementsException, QualityException

_LOGGER = logging.getLogger("bedboss")


def make_bigbed(
    bed: Union[str, RegionSet],
    output_path: str,
    genome: str,
    rfg_config: Union[str, Path] = None,
) -> None:
    """
    Generate bigBed file for the BED file.

    :param bed: path to the BED file, or RegionSet object from Rust
    :param genome: reference genome (e.g. hg38, mm10, etc.)
    :param output_path: full path to the output bigBed file

    :return: None
    """

    try:
        chrom_sizes = get_chrom_sizes(genome=genome, rfg_config=rfg_config)
    except MissingGenomeError:
        raise BedBossException("Could not find Genome in refgenie. Skipping...")

    if isinstance(bed, str):
        bed = RegionSet(bed)
    elif not isinstance(bed, RegionSet):
        raise BedBossException("Invalid bed object. Must be a path or RegionSet.")

    try:
        if not chrom_sizes:
            raise BedBossException("Chrom sizes not found. Skipping...")
        bed.to_bigbed(output_path, chrom_sizes)
    except BaseException as err:
        raise BedBossException(
            f"Failed to generate bigBed file for {output_path}: Error: {err}"
        )

    _LOGGER.info(f"BigBed file generated: {output_path}")


def make_bed(
    input_file: str,
    input_type: str,
    output_path: str,
    genome: str,
    narrowpeak: bool = False,
    rfg_config: str = None,
    chrom_sizes: str = None,
    pm: pypiper.PipelineManager = None,
) -> Tuple[str, str, RegionSet]:
    """
    Convert the input file to BED format by construct the command based
    on input file type and execute the command.

    :param input_file: path to the input file
    :param input_type: a [bigwig|bedgraph|bed|bigbed|wig] file that will be
                        converted into BED format
    :param output_path: path to the output folder, logs will be saved (BED files will be cached)
    :param genome: reference genome
    :param narrowpeak: whether the regions are narrow (transcription factor
                        implies narrow, histone mark implies broad peaks)
    :param rfg_config: file path to the genome config file
    :param chrom_sizes: a full path to the chrom.sizes required for the
                        wig files conversion
    :param pm: pypiper object

    :return: path to the BED file
    """
    _LOGGER.info(f"Processing {input_file} file in bedmaker...")

    bbclient = BBClient()

    input_type = input_type.lower()
    if input_type not in [member.value for member in InputTypes]:
        raise BedBossException(
            f"Invalid input type: {input_type}. "
            f"Supported types: {', '.join([k.value for k in InputTypes])}"
        )

    if not pm:
        pm = pypiper.PipelineManager(
            name="bedmaker",
            outfolder=os.path.join(os.path.dirname(output_path), "bedmaker_logs"),
            recover=True,
            multi=True,
        )
        pm_clean = True
    else:
        pm_clean = False

    temp_bed_path = os.path.splitext(output_path)[0]

    file_base_name = os.path.basename(input_file)
    input_extension = os.path.splitext(file_base_name)[1]

    width = "bdgbroadcall" if not narrowpeak else "bdgpeakcall"

    # creat cmd to run that convert non bed file to bed file
    if input_type == InputTypes.BED.value:
        try:
            bed_obj = bbclient.add_bed_to_cache(input_file)
            bed_id = bed_obj.identifier
            output_path = bbclient.seek(bed_id)
        except BaseException as e:
            raise BedBossException(f"File not found: {input_file} Error: {e}")

    else:
        _LOGGER.info(f"Converting {input_file} to BED format")

        # Use the gzip and shutil modules to produce temporary unzipped files
        if input_extension == ".gz":
            temp_input_file = os.path.join(
                os.path.dirname(output_path),
                os.path.splitext(file_base_name)[0],
            )
            with gzip.open(input_file, "rb") as f_in:
                with open(temp_input_file, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            pm.clean_add(temp_input_file)

        # creating cmd for bedGraph files
        if input_type == InputTypes.BED_GRAPH.value:
            if not is_command_callable("macs2"):
                raise RequirementsException(
                    "To convert bedGraph file You must first install "
                    "macs2 and add it to your PATH. "
                    "Instruction: "
                    "https://pypi.org/project/MACS2/"
                )
            else:
                cmd = BEDGRAPH_TEMPLATE.format(
                    input=input_file,
                    output=temp_bed_path,
                    width=width,
                )

        # creating cmd for bigWig files
        elif input_type == InputTypes.BIG_WIG.value:
            if not is_command_callable("bigWigToBedGraph"):
                raise RequirementsException(
                    "To convert bigWig file You must first install "
                    "bigWigToBedGraph and add it to your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigWig.html"
                )
            else:
                cmd = BIGWIG_TEMPLATE.format(
                    input=input_file,
                    output=temp_bed_path,
                    width=width,
                )
        # creating cmd for wig files
        elif input_type == InputTypes.WIG.value:
            if not chrom_sizes:
                chrom_sizes = get_chrom_sizes(genome=genome, rfg_config=rfg_config)

            # define a target for temporary bw files
            temp_target = os.path.join(output_path, file_base_name + ".bw")
            if not is_command_callable("wigToBigWig"):
                raise RequirementsException(
                    "To convert wig file You must first install "
                    "wigToBigWig and add it in your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigWig.html"
                )
            else:
                pm.clean_add(temp_target)
                cmd1 = WIG_TEMPLATE.format(
                    input=input_file,
                    intermediate_bw=temp_target,
                    chrom_sizes=chrom_sizes,
                    width=width,
                )

            if not is_command_callable("bigWigToBedGraph"):
                raise RequirementsException(
                    "To convert bigWig file You must first install "
                    "bigWigToBedGraph and add it in your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigWig.html"
                )
            else:
                cmd2 = BIGWIG_TEMPLATE.format(
                    input=temp_target,
                    output=temp_bed_path,
                    width=width,
                )
                cmd = [cmd1, cmd2]
        # creating cmd for bigBed files
        elif input_type == InputTypes.BIG_BED.value:
            if not is_command_callable(BIGBED_TO_BED_PROGRAM):
                raise RequirementsException(
                    "To convert bigBed file You must first install "
                    "bigBedToBed and add it in your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigBed.html"
                )
            else:
                cmd = BIGBED_TEMPLATE.format(input=input_file, output=temp_bed_path)

        else:
            raise NotImplementedError(f"'{input_type}' format is not supported")

        pm.run(cmd, temp_bed_path, nofail=False)

        bed_obj = bbclient.add_bed_to_cache(input_file)
        bed_id = bed_obj.identifier
        output_path = bbclient.seek(bed_id)

    pm._cleanup()
    if pm_clean:
        pm.stop_pipeline()

    _LOGGER.info(
        f"Bed output file: {output_path}. BEDmaker: File processed successfully."
    )

    return output_path, bed_id, bed_obj


def make_all(
    input_file: str,
    input_type: str,
    output_path: str,
    genome: str,
    rfg_config: str = None,
    chrom_sizes: str = None,
    narrowpeak: bool = False,
    check_qc: bool = True,
    lite: bool = False,
    pm: pypiper.PipelineManager = None,
) -> BedMakerOutput:
    """
    Maker of bed and bigbed files.

    Pipeline to convert supported file formats into
    BED format and bigBed format. Currently supported formats*:
        - bedGraph
        - bigBed
        - bigWig
        - wig
        - bed
    :param input_file: path to the input file
    :param input_type: a [bigwig|bedgraph|bed|bigbed|wig] file that will be
                       converted into BED format
    :param output_path: path to the output folder, where bigbed and logs will be saved
    :param genome: reference genome
    :param rfg_config: file path to the genome config file
    :param chrom_sizes: a full path to the chrom.sizes required for the
                        bedtobigbed conversion
    :param narrowpeak: whether the regions are narrow (transcription factor
                       implies narrow, histone mark implies broad peaks)
    :param check_qc: run quality control during bedmaking
    :param lite: run the pipeline in lite mode (without producing bigBed files)
    :param pm: pypiper object

    :return: dict with generated bed metadata - BedMakerOutput object:
        {
            "bed_compliance": bed_compliance. e.g. bed3+0
            "data_format": data_format. e.g. narrowpeak, broadpeak
            "bed_file": path to the bed file
            "bigbed_file": path to the bigbed file
            "bed_digest": bed_digest
        }
    """
    if not pm:
        pm = pypiper.PipelineManager(
            name="bedmaker",
            outfolder=os.path.join(output_path, "bedmaker_logs"),
            recover=True,
            multi=True,
        )
        pm_clean = True
    else:
        pm_clean = False
    output_bed, bed_id, bed_obj = make_bed(
        input_file=input_file,
        input_type=input_type,
        output_path=output_path,
        genome=genome,
        narrowpeak=narrowpeak,
        rfg_config=rfg_config,
        chrom_sizes=chrom_sizes,
        pm=pm,
    )
    bed_classification = get_bed_classification(output_bed)
    if check_qc:
        try:
            file_size = os.path.getsize(output_bed)
            if file_size >= MAX_FILE_SIZE:
                raise QualityException(
                    f"File size is larger than {MAX_FILE_SIZE} bytes. File size= {file_size} bytes."
                )

            mean_region_width = bed_obj.mean_region_width()
            if mean_region_width < MIN_REGION_WIDTH:
                raise QualityException(
                    f"Mean region width is less than {MIN_REGION_WIDTH} bp. File mean region width= {mean_region_width} bp."
                )
            number_of_regions = len(bed_obj)
            if number_of_regions > MAX_REGION_NUMBER:
                raise QualityException(
                    f"Number of regions is greater than {MAX_REGION_NUMBER}. File number of regions= {number_of_regions}."
                )
        except QualityException as e:
            raise QualityException(
                f"Quality control failed for {output_path}. Error: {e}"
            )

        _LOGGER.info(f"File ({output_bed}) has passed Quality Control!")

    if lite:
        _LOGGER.info("Skipping bigBed generation due to lite mode.")
        output_bigbed = None
    else:
        try:
            bigbed_folder_path = os.path.join(
                output_path, BIGBED_FOLDER_NAME, bed_id[0], bed_id[1]
            )
            os.makedirs(bigbed_folder_path, exist_ok=True)
            output_bigbed = os.path.join(bigbed_folder_path, f"{bed_id}.bigBed")

            make_bigbed(
                bed=bed_obj,
                output_path=output_bigbed,
                genome=genome,
                rfg_config=rfg_config,
            )
        except BedBossException:
            output_bigbed = None
    if pm_clean:
        pm.stop_pipeline()

    _LOGGER.info(f"Bed output file: {output_bed}")
    _LOGGER.info(f"BigBed output file: {output_bigbed}")

    return BedMakerOutput(
        bed_object=bed_obj,
        bed_file=output_bed,
        bigbed_file=os.path.abspath(output_bigbed) if output_bigbed else None,
        bed_digest=bed_id,
        bed_compliance=bed_classification.bed_compliance,
        compliant_columns=bed_classification.compliant_columns,
        non_compliant_columns=bed_classification.non_compliant_columns,
        data_format=bed_classification.data_format,
    )



import peppy
import pytest

from bedboss.utils import standardize_pep


# @pytest.mark.skip(reason="Not for automatic testing")
@pytest.mark.parametrize(
    "registry_path",
    [
        "bedbase/gse274894:samples",
        "bedbase/gse275349:samples",
        "bedbase/gse262920:samples",
        "bedbase/gse236101:samples",
        "bedbase/gse254365:samples",
    ],
)
@pytest.mark.parametrize(
    "model",
    ["BEDBASE", "ENCODE"],
)
def test_standardize_pep(registry_path, model):
    pep = peppy.Project.from_pephub(registry_path)
    standardize_pep(pep, model=model)
    assert True



import logging
import os
from typing import Literal, Union

import pypiper

import peppy
from bbconf import BedBaseAgent
from bbconf.db_utils import GeoGseStatus, GeoGsmStatus
from geniml.exceptions import GenimlBaseError
from pephubclient import PEPHubClient
from pephubclient.helpers import MessageHandler
from pephubclient.models import SearchReturnModel
from sqlalchemy import and_, select
from sqlalchemy.orm import Session

from bedboss.bbuploader.constants import (
    DEFAULT_GEO_TAG,
    FILE_FOLDER_NAME,
    PKG_NAME,
    STATUS,
)
from bedboss.bbuploader.metadata_extractor import find_cell_line, find_assay
from bedboss.bbuploader.models import (
    BedBossMetadata,
    BedBossRequired,
    ProjectProcessingStatus,
    BedBossMetadataSeries,
)
from bedboss.bbuploader.utils import (
    create_gsm_sub_name,
    build_gse_identifier,
    middle_underscored,
)
from bedboss.bedboss import run_all
from bedboss.bedbuncher.bedbuncher import run_bedbuncher
from bedboss.const import MAX_FILE_SIZE
from bedboss.exceptions import BedBossException, QualityException
from bedboss.skipper import Skipper
from bedboss.refgenome_validator.main import ReferenceValidator
from bedboss.utils import (
    calculate_time,
    download_file,
    standardize_genome_name,
    run_initial_qc,
)
from bedboss.utils import standardize_pep as pep_standardizer
from bedboss.bedstat.r_service import RServiceManager
from bedboss._version import __version__

_LOGGER = logging.getLogger(PKG_NAME)
_LOGGER.setLevel(logging.DEBUG)


reference_validator = ReferenceValidator()


@calculate_time
def upload_all(
    bedbase_config: str,
    outfolder: str = os.getcwd(),
    geo_tag: str = DEFAULT_GEO_TAG,
    start_date: str = None,
    end_date: str = None,
    search_limit: int = 10,
    search_offset: int = 0,
    download_limit: int = 100,
    genome: str = None,
    create_bedset: bool = True,
    preload=True,
    rerun: bool = False,
    run_skipped: bool = False,
    run_failed: bool = True,
    standardize_pep: bool = False,
    use_skipper=True,
    reinit_skipper=False,
    overwrite=False,
    overwrite_bedset=False,
    lite=False,
):
    """
    This is main function that is responsible for processing bed files from PEPHub.

    :param outfolder: working directory, where files will be downloaded, processed and statistics will be saved
    :param bedbase_config: path to bedbase configuration file
    :param geo_tag: GEO tag to use when loading projects from PEPHub ('samples' or 'series')
    :param start_date: The earliest date when opep was updated [Default: 2000/01/01]
    :param end_date: The latest date when opep was updated [Default: today's date]
    :param search_limit: limit of projects to be searched
    :param search_offset: offset of projects to be searched
    :param download_limit: limit of GSE projects to be downloaded (used for testing purposes) [Default: 100]
    :param genome: reference genome [Default: None] (e.g. hg38) - if None, all genomes will be processed
    :param create_bedset: create bedset from bed files
    :param preload: pre - download files to the local folder (used for faster reproducibility)
    :param rerun: rerun processing of the series. Used in logging system. If you want to reupload file use overwrite
    :param run_skipped: rerun files that were skipped. Used in logging system. If you want to reupload file use overwrite
    :param run_failed: rerun failed files. Used in logging system. If you want to reupload file use overwrite
    :param standardize_pep: standardize pep metadata using BEDMS
    :param use_skipper: use skipper to skip already processed logged locally. Skipper creates local log of processed
        and failed files.
    :param reinit_skipper: reinitialize skipper, if set to True, skipper will be reinitialized and all logs files will be cleaned
    :param lite: lite mode, where skipping statistic processing for memory optimization and time saving
    """

    phc = PEPHubClient()
    os.makedirs(outfolder, exist_ok=True)

    _LOGGER.info(f"Initializing BedBaseAgent with config: '{bedbase_config}'")
    bbagent = BedBaseAgent(config=bedbase_config, init_ml=not lite)
    _LOGGER.info(f"BedBaseAgent initialized (ML enabled: {not lite})")

    genome = standardize_genome_name(genome, reference_validator=reference_validator)
    if genome:
        _LOGGER.info(f"Filtering for genome: '{genome}'")

    pep_annotation_list = find_peps(
        geo_tag=geo_tag,
        start_date=start_date,
        end_date=end_date,
        limit=search_limit,
        offset=search_offset,
        phc=phc,
    )

    _LOGGER.info(
        f"parameters: start_date={start_date}, end_date={end_date}, search_limit={search_limit}, search_offset={search_offset}"
    )
    _LOGGER.info(f"found {pep_annotation_list.count} projects")

    count = 0
    total_projects = len(pep_annotation_list.results)

    pm_out_folder = os.path.join(os.path.abspath(outfolder), "pipeline_manager")
    _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
    pm = pypiper.PipelineManager(
        name="bedboss-pipeline",
        outfolder=pm_out_folder,
        version=__version__,
        recover=True,
    )

    if not lite:
        _LOGGER.info("Initializing R service for statistics")
        r_service = RServiceManager()
    else:
        _LOGGER.info("Lite mode: R service disabled")
        r_service = None

    for gse_pep in pep_annotation_list.results:
        count += 1
        gse_id = build_gse_identifier(gse_pep.name, geo_tag)
        with Session(bbagent.config.db_engine.engine) as session:
            MessageHandler.print_success(f"{'##' * 30}")
            MessageHandler.print_success(
                f"#### Processing: '{gse_id}'. #### Processing {count} / {total_projects}. ####"
            )

            gse_status = session.scalar(
                select(GeoGseStatus).where(GeoGseStatus.gse == gse_id)
            )

            if gse_status:
                if gse_status.status == STATUS.SUCCESS and not rerun:
                    _LOGGER.info(
                        f"Skipping: '{gse_id}' - already processed and rerun set to false"
                    )
                    continue

                elif (
                    gse_status.status == STATUS.FAIL
                    or gse_status.status == STATUS.PROCESSING
                ) and not run_failed:
                    _LOGGER.info("Reprocessing of failed set to false, exiting.")
                    continue

                elif (
                    gse_status.status == STATUS.SKIPPED
                    or gse_status.status == STATUS.PARTIAL
                ) and not run_skipped:
                    _LOGGER.info(
                        f"Run skipped files set to false, exiting. GSE: {gse_id}"
                    )
                    continue

            else:
                gse_status = GeoGseStatus(gse=gse_id, status=STATUS.PROCESSING)
                session.add(gse_status)
                session.commit()

            try:
                upload_result = _upload_gse(
                    gse=gse_pep.name,
                    bedbase_config=bbagent,
                    outfolder=outfolder,
                    geo_tag=geo_tag,
                    create_bedset=create_bedset,
                    genome=genome,
                    sa_session=session,
                    gse_status_sa_model=gse_status,
                    standardize_pep=standardize_pep,
                    rerun=rerun,
                    use_skipper=use_skipper,
                    reinit_skipper=reinit_skipper,
                    preload=preload,
                    overwrite=overwrite,
                    overwrite_bedset=overwrite_bedset,
                    lite=lite,
                    r_service=r_service,
                    pm=pm,
                )
            except Exception as err:
                _LOGGER.error(
                    f"Processing of '{gse_pep.name}' failed with error: {err}"
                )
                gse_status.status = STATUS.FAIL
                gse_status.error = str(err)
                session.commit()
                continue

            status_parser(gse_status, upload_result)
            session.commit()

            if count >= download_limit:
                break

    pm.stop_pipeline()

    return None


def process_pep_sample(
    bed_sample: peppy.Sample,
    geo_tag: str = DEFAULT_GEO_TAG,
) -> BedBossRequired:
    """
    Process pep sample that contains bed file. Download bed file and compose BedBossRequired data model
        that contains all required bed file metadata (e.g. reference genome, type, organism...)

    :param bed_sample: peppy sample with bed file url
    :param geo_tag: GEO tag to use when loading projects from PEPHub ('samples' or 'series')
    :return: BedBossRequired {sample_name: str,
                              gse: str,
                              gsm: str,
                              file_path: str,
                              ref_genome:str,
                              ...
                                }
    """
    _LOGGER.debug(
        f"Standardizing metadata for: '{bed_sample.sample_name}' . GSE: {bed_sample.gse}"
    )

    if geo_tag == "series":
        project_metadata = BedBossMetadataSeries(**bed_sample.to_dict())

        description_text = f"{bed_sample.sample_name} {project_metadata.description}"
        if not project_metadata.assay or project_metadata.assay == "OTHER":
            predicted_assay = find_assay(description_text)
            if predicted_assay:
                project_metadata.assay = predicted_assay

        if not project_metadata.cell_line:
            predicted_cell_line = find_cell_line(description_text)
            if predicted_cell_line:
                project_metadata.cell_line = predicted_cell_line

        return BedBossRequired(
            sample_name=bed_sample.sample_name,
            file_path=bed_sample.file_url,
            ref_genome="undefined",
            type="bed",
            narrowpeak=False,
            pep=project_metadata,
            title=middle_underscored(bed_sample.get("sample_name", "")),
        )
    elif geo_tag == "samples":
        if bed_sample.type == "NARROWPEAK":
            file_type = "bed"
            is_narrowpeak = True
        elif bed_sample.type == "BROADPEAK":
            file_type = "bed"
            is_narrowpeak = False
        else:
            file_type = "bed"
            is_narrowpeak = False

        project_metadata = BedBossMetadata(**bed_sample.to_dict())

        description_text = f"{bed_sample.sample_name} {project_metadata.description}"

        if not project_metadata.assay or project_metadata.assay == "OTHER":
            predicted_assay = find_assay(description_text)
            if predicted_assay:
                project_metadata.assay = predicted_assay

        if not project_metadata.cell_line:
            predicted_cell_line = find_cell_line(description_text)
            if predicted_cell_line:
                project_metadata.cell_line = predicted_cell_line

        return BedBossRequired(
            sample_name=bed_sample.sample_name,
            file_path=bed_sample.file_url,
            ref_genome=(
                bed_sample.ref_genome.strip()
                if bed_sample.ref_genome
                else bed_sample.ref_genome
            ),
            type=file_type,
            narrowpeak=is_narrowpeak,
            pep=project_metadata,
            title=bed_sample.get("sample_title", ""),
        )
    else:
        raise BedBossException(
            f"Unsupported geo_tag: '{geo_tag}'. Please use 'samples' or 'series'."
        )


def get_pep(
    namespace: str,
    name: str,
    tag: str,
    phc: PEPHubClient = None,
):
    """
    Retrieve PEP from PEPHub

    :param namespace: namespace of the project
    :param name: name of the project
    :param tag: tag of the project
    :param phc: PEPHubClient instance
    :return: ProjectModel
    """
    if not phc:
        phc = PEPHubClient()
    return phc.load_project(f"{namespace}/{name}:{tag}")


def find_peps(
    start_date: str = "2000/01/01",
    end_date: str = None,
    filter_by: Literal["submission_date", "last_update_date"] = "submission_date",
    limit: int = 10000,
    offset: int = 0,
    geo_tag: str = DEFAULT_GEO_TAG,
    phc: PEPHubClient = None,
) -> SearchReturnModel:
    """
    Retrieve list of PEPs from 'bedbase' namespace in certain time period

    :param start_date: earliest date when opep was updated [Default: 2000/01/01]
    :param end_date: latest date when opep was updated [Default: today's date]
    :param filter_by: filter by submission date [Default: submission_date] Option: [submission_date, last_update_date]
    :param limit: limit of projects to be searched
    :param offset: offset of projects to be searched
    :param geo_tag: GEO tag to use when loading projects from PEPHub ('samples' or 'series')
    :param phc: PEPHubClient instance
    :return SearchReturnModel: {count: int
                                limit: int
                                offset: int
                                items: List[ProjectAnnotationModel]
                            }
    """
    if not phc:
        phc = PEPHubClient()
    _LOGGER.info(
        f"Searching PEPHub for projects (namespace='bedbase', tag='{geo_tag}', "
        f"dates={start_date} to {end_date}, limit={limit}, offset={offset})"
    )
    return phc.find_project(
        namespace="bedbase",
        tag=geo_tag,
        limit=limit,
        offset=offset,
        filter_by=filter_by,
        start_date=start_date,
        end_date=end_date,
    )


@calculate_time
def upload_gse(
    gse: str,
    bedbase_config: Union[str, BedBaseAgent],
    outfolder: str = os.getcwd(),
    geo_tag: str = DEFAULT_GEO_TAG,
    create_bedset: bool = True,
    genome: str = None,
    preload: bool = True,
    rerun: bool = False,
    run_skipped: bool = False,
    run_failed: bool = True,
    standardize_pep: bool = False,
    use_skipper=True,
    reinit_skipper=False,
    overwrite=False,
    overwrite_bedset=True,
    lite=False,
):
    """
    Upload bed files from GEO series to BedBase

    :param gse: GEO series number
    :param bedbase_config: path to bedbase configuration file, or bbagent object
    :param outfolder: working directory, where files will be downloaded, processed and statistics will be saved
    :param geo_tag: GEO tag to use when loading projects from PEPHub ('samples' or 'series')
    :param create_bedset: create bedset from bed files
    :param genome: reference genome to upload to database. If None, all genomes will be processed
    :param preload: pre - download files to the local folder (used for faster reproducibility)
    :param rerun: rerun processing of the series
    :param run_skipped: rerun files that were skipped
    :param run_failed: rerun failed files
    :param standardize_pep: standardize pep metadata using BEDMS
    :param use_skipper: use skipper to skip already processed logged locally. Skipper creates local log of processed
        and failed files.
    :param reinit_skipper: reinitialize skipper, if set to True, skipper will be reinitialized and all logs files will be cleaned
    :param overwrite: overwrite existing bedfiles
    :param overwrite_bedset: overwrite existing bedset
    :param lite: lite mode, where skipping statistic processing for memory optimization and time saving

    :return: None
    """
    _LOGGER.info(f"Initializing BedBaseAgent with config: '{bedbase_config}'")
    bbagent = BedBaseAgent(config=bedbase_config, init_ml=not lite)
    _LOGGER.info(f"BedBaseAgent initialized (ML enabled: {not lite})")
    gse_id = build_gse_identifier(gse, geo_tag)

    with Session(bbagent.config.db_engine.engine) as session:
        _LOGGER.info(f"Processing: '{gse_id}'")

        gse_status = session.scalar(
            select(GeoGseStatus).where(GeoGseStatus.gse == gse_id)
        )
        if gse_status:
            if gse_status.status == STATUS.SUCCESS and not rerun:
                _LOGGER.info(
                    f"Skipping: '{gse_id}' - already processed and rerun set to false"
                )
                exit()

            elif (
                gse_status.status == STATUS.FAIL
                or gse_status.status == STATUS.PROCESSING
            ) and not run_failed:
                _LOGGER.info("Reprocessing of failed set to false, exiting.")
                exit()

            elif (
                gse_status.status == STATUS.SKIPPED
                or gse_status.status == STATUS.PARTIAL
            ) and not run_skipped:
                _LOGGER.info(f"Run skipped files set to false, exiting. GSE: {gse_id}")
                exit()

        else:
            gse_status = GeoGseStatus(gse=gse_id, status=STATUS.FAIL)
            session.add(gse_status)
            session.commit()

        try:
            upload_result = _upload_gse(
                gse=gse,
                bedbase_config=bbagent,
                outfolder=outfolder,
                geo_tag=geo_tag,
                create_bedset=create_bedset,
                genome=genome,
                sa_session=session,
                gse_status_sa_model=gse_status,
                standardize_pep=standardize_pep,
                preload=preload,
                overwrite=overwrite,
                rerun=rerun,
                overwrite_bedset=overwrite_bedset,
                use_skipper=use_skipper,
                reinit_skipper=reinit_skipper,
                lite=lite,
            )
        except Exception as e:
            _LOGGER.error(f"Processing of '{gse_id}' failed with error: {e}")
            gse_status.status = STATUS.FAIL
            gse_status.error = str(e)
            session.commit()
            exit()

        status_parser(gse_status, upload_result)

        _LOGGER.info(f"Processing of '{gse_id}' is finished with success!")
        session.commit()


def status_parser(
    gse_status: GeoGseStatus, upload_result: ProjectProcessingStatus
) -> None:
    """
    Update status in SA object

    :param gse_status: gse status of project (Sqlalchemy object)
    :param upload_result: project processing status (status object)
    """

    gse_status.number_of_files = upload_result.number_of_samples
    gse_status.number_of_success = upload_result.number_of_processed
    gse_status.number_of_skips = upload_result.number_of_skipped
    gse_status.number_of_fails = upload_result.number_of_failed
    if upload_result.number_of_samples == upload_result.number_of_processed:
        gse_status.status = STATUS.SUCCESS
    elif upload_result.number_of_skipped == upload_result.number_of_samples:
        gse_status.status = STATUS.SKIPPED
    elif upload_result.number_of_skipped > 0 and upload_result.number_of_failed == 0:
        gse_status.status = STATUS.PARTIAL
    else:
        gse_status.status = STATUS.FAIL


def _upload_gse(
    gse: str,
    bedbase_config: Union[str, BedBaseAgent],
    outfolder: str = os.getcwd(),
    geo_tag: str = DEFAULT_GEO_TAG,
    create_bedset: bool = True,
    genome: str = None,
    sa_session: Session = None,
    gse_status_sa_model: GeoGseStatus = None,
    standardize_pep: bool = False,
    rerun: bool = False,
    overwrite: bool = False,
    overwrite_bedset: bool = False,
    use_skipper: bool = True,
    reinit_skipper: bool = False,
    preload: bool = True,
    lite=False,
    max_file_size: int = 20 * 1000000,
    r_service: RServiceManager = None,
    pm: pypiper.PipelineManager = None,
) -> ProjectProcessingStatus:
    """
    Upload bed files from GEO series to BedBase

    :param gse: GEO series number
    :param bedbase_config: path to bedbase configuration file, or bbagent object
    :param outfolder: working directory, where files will be downloaded, processed and statistics will be saved
    :param geo_tag: GEO tag to use when loading projects from PEPHub ('samples' or 'series')
    :param create_bedset: create bedset from bed files
    :param genome: reference genome to upload to database. If None, all genomes will be processed
    :param sa_session: opened session to the database
    :param gse_status_sa_model: sqlalchemy model for project status
    :param standardize_pep: standardize pep metadata using BEDMS
    :param rerun: rerun processing of the series
    :param overwrite: overwrite existing bedfiles
    :param overwrite_bedset: overwrite existing bedset
    :param use_skipper: use skipper to skip already processed logged locally. Skipper creates local log of processed
        and failed files.
    :param reinit_skipper: reinitialize skipper, if set to True, skipper will be reinitialized and all logs will be
    :param preload: pre - download files to the local folder (used for faster reproducibility)
    :param lite: lite mode, where skipping statistic processing for memory optimization and time saving
    :param max_file_size: maximum file size in bytes. Default: 20MB
    :param pypiper.PipelineManager pm: pypiper object
    :param r_service: RServiceManager object
    :return: None
    """
    if isinstance(bedbase_config, str):
        bedbase_config = BedBaseAgent(config=bedbase_config)
    if genome:
        genome = standardize_genome_name(
            genome, reference_validator=reference_validator
        )

    gse_id = build_gse_identifier(gse, geo_tag)

    phc = PEPHubClient()
    os.makedirs(outfolder, exist_ok=True)

    _LOGGER.info(f"Loading project from PEPHub: 'bedbase/{gse}:{geo_tag}'")
    project = phc.load_project(f"bedbase/{gse}:{geo_tag}")
    _LOGGER.info(f"Loaded project with {len(project.samples)} samples")

    if standardize_pep:
        project = pep_standardizer(project)

    project_status = ProjectProcessingStatus(number_of_samples=len(project.samples))
    uploaded_files = []
    gse_status_sa_model.number_of_files = len(project.samples)
    sa_session.commit()

    total_sample_number = len(project.samples)

    if use_skipper:
        skipper_obj = Skipper(output_path=outfolder, name=gse_id)
        if reinit_skipper:
            skipper_obj.reinitialize()
        _LOGGER.info(f"Skipper initialized for: '{gse_id}'")
    else:
        skipper_obj = None

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(outfolder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    if not lite and not r_service:
        r_service = RServiceManager()
    elif lite:
        r_service = None
    else:
        r_service = r_service

    for counter, project_sample in enumerate(project.samples):
        _LOGGER.info(f">> Processing {counter+1} / {total_sample_number}")
        sample_gsm = project_sample.get("sample_geo_accession", "").lower()
        sample_sample_name = project_sample.get("sample_name", "").lower()

        if skipper_obj:
            is_processed = skipper_obj.is_processed(
                f"{sample_gsm}_{sample_sample_name}"
            )
            if is_processed:
                _LOGGER.info(
                    f"Skipping: '{sample_gsm}_{sample_sample_name}' - already processed"
                )
                uploaded_files.append(is_processed)
                continue

        required_metadata = process_pep_sample(
            bed_sample=project_sample,
            geo_tag=geo_tag,
        )

        sample_status = sa_session.scalar(
            select(GeoGsmStatus).where(
                and_(
                    GeoGsmStatus.sample_name == required_metadata.sample_name,
                    GeoGsmStatus.gse_status_id == gse_status_sa_model.id,
                )
            )
        )

        if not sample_status:
            sample_status = GeoGsmStatus(
                gse_status_mapper=gse_status_sa_model,
                sample_name=required_metadata.sample_name,
                gsm=sample_gsm,
                status=STATUS.PROCESSING,
            )
            sa_session.add(sample_status)
            sa_session.commit()
        else:
            if sample_status.status == STATUS.SUCCESS and not rerun:
                _LOGGER.info(
                    f"Skipping: '{required_metadata.sample_name}' - already processed"
                )
                uploaded_files.append(sample_status.bed_id)
                project_status.number_of_processed += 1
                continue

        sample_status.genome = required_metadata.ref_genome
        # to upload files only with a specific genome
        if genome:
            if required_metadata.ref_genome != genome:
                _LOGGER.info(
                    f"Skipping: '{required_metadata.sample_name}' - genome mismatch. Expected: '{genome}' Found: '{required_metadata.ref_genome}'"
                )
                sample_status.status = STATUS.SKIPPED

                sa_session.commit()
                project_status.number_of_skipped += 1

                continue

        _LOGGER.info(
            f"Processing global_sample_id: '{required_metadata.pep.global_sample_id}' file: '{required_metadata.sample_name}' gse: '{gse}', geo_tag: '{geo_tag}'"
        )
        sample_status.status = STATUS.PROCESSING
        sa_session.commit()

        sample_status.file_size = project_sample.get("file_size", 0)
        sample_status.source_submission_date = project_sample.get(
            "sample_submission_date", None
        )

        try:
            if int(project_sample.get("file_size") or 0) > max_file_size:
                raise QualityException(
                    f"File size is too big. {int(project_sample.get('file_size', 0)) / 1000000} MB"
                )

            # to speed up the process, we can run initial QC on the file
            qc_file_size = run_initial_qc(project_sample.file_url)
            if qc_file_size > 0:
                sample_status.file_size = qc_file_size
        except QualityException as err:
            _LOGGER.error(f"Processing of '{sample_gsm}' failed with error: {str(err)}")
            sample_status.status = STATUS.FAIL
            sample_status.error = str(err)
            if err.file_size > 0:
                sample_status.file_size = min(
                    err.file_size, MAX_FILE_SIZE
                )  # we need to limit file size to MAX_FILE_SIZE for DB storage
            project_status.number_of_failed += 1

            if skipper_obj:
                skipper_obj.add_failed(
                    f"{sample_gsm}_{sample_sample_name}", f"Error: {str(err)}"
                )
            sa_session.commit()
            continue

        if preload:
            gsm_folder = create_gsm_sub_name(sample_gsm)
            files_path = os.path.join(outfolder, FILE_FOLDER_NAME, gsm_folder)
            os.makedirs(files_path, exist_ok=True)
            file_abs_path = os.path.abspath(
                os.path.join(files_path, project_sample.file)
            )
            _LOGGER.info(f"Downloading file to: '{file_abs_path}'")
            download_file(project_sample.file_url, file_abs_path, no_fail=True)
        else:
            file_abs_path = required_metadata.file_path

        try:
            original_genome = required_metadata.ref_genome
            # This code will standardize or predict genome if not provided
            required_metadata.ref_genome = standardize_genome_name(
                required_metadata.ref_genome,
                file_abs_path,
                reference_validator=reference_validator,
            )
            if original_genome != required_metadata.ref_genome:
                _LOGGER.info(
                    f"Genome standardized: '{original_genome}' -> '{required_metadata.ref_genome}'"
                )

            _LOGGER.info(
                f"Starting bed processing for '{required_metadata.sample_name}' (genome: {required_metadata.ref_genome})"
            )
            file_digest = run_all(
                name=required_metadata.title,
                input_file=file_abs_path,
                input_type=required_metadata.type,
                outfolder=os.path.join(outfolder, "outputs"),
                genome=required_metadata.ref_genome,
                bedbase_config=bedbase_config,
                narrowpeak=required_metadata.narrowpeak,
                other_metadata=required_metadata.pep.model_dump(),
                upload_pephub=True,
                upload_s3=True,
                upload_qdrant=True,
                force_overwrite=overwrite,
                lite=lite,
                pm=pm,
                r_service=r_service,
                reference_genome_validator=reference_validator,
            )
            _LOGGER.info(
                f"Successfully processed '{required_metadata.sample_name}' -> digest: {file_digest}"
            )
            uploaded_files.append(file_digest)
            if skipper_obj:
                skipper_obj.add_processed(
                    f"{sample_gsm}_{sample_sample_name}", file_digest
                )
            sample_status.status = STATUS.SUCCESS
            sample_status.bed_id = file_digest
            project_status.number_of_processed += 1

        except BedBossException as exc:
            _LOGGER.error(f"Processing of '{sample_gsm}' failed with error: {str(exc)}")
            sample_status.status = STATUS.FAIL
            sample_status.error = str(exc)
            project_status.number_of_failed += 1

            if skipper_obj:
                skipper_obj.add_failed(
                    f"{sample_gsm}_{sample_sample_name}", f"Error: {str(exc)}"
                )

        except GenimlBaseError as exc:
            _LOGGER.error(f"Processing of '{sample_gsm}' failed with error: {str(exc)}")
            sample_status.status = STATUS.FAIL
            sample_status.error = str(exc)
            project_status.number_of_failed += 1

            if skipper_obj:
                skipper_obj.add_failed(
                    f"{sample_gsm}_{sample_sample_name}", f"Error: {str(exc)}"
                )

        sa_session.commit()

    if create_bedset and uploaded_files:
        _LOGGER.info(f"Creating bedset for: '{gse_id}'")

        experiment_metadata = project.config.get("experiment_metadata", {})

        # Check if bedset already exists and merge bed IDs
        combined_bed_ids = list(uploaded_files)
        try:
            existing_bedset = bedbase_config.bedset.get(gse)
            if existing_bedset and existing_bedset.bed_ids:
                _LOGGER.info(
                    f"Bedset '{gse}' already exists with {len(existing_bedset.bed_ids)} files. Merging with {len(uploaded_files)} new files."
                )
                # Merge existing and new bed IDs, avoiding duplicates
                existing_ids = set(existing_bedset.bed_ids)
                for bed_id in uploaded_files:
                    if bed_id not in existing_ids:
                        existing_ids.add(bed_id)
                combined_bed_ids = list(existing_ids)
                _LOGGER.info(
                    f"Combined bedset will have {len(combined_bed_ids)} files."
                )
        except Exception:
            _LOGGER.debug(f"Bedset '{gse}' does not exist yet. Creating new bedset.")

        run_bedbuncher(
            bedbase_config=bedbase_config,
            record_id=gse,
            bed_set=combined_bed_ids,
            output_folder=os.path.join(outfolder, "outputs"),
            name=gse,
            description=project.description,
            heavy=False,  # TODO: set to False because can't handle bedset > 10 files
            upload_pephub=True,
            upload_s3=True,
            no_fail=True,
            force_overwrite=overwrite_bedset,
            lite=lite,
            annotation={
                "summary": experiment_metadata.get("series_summary", ""),
                "author": ", ".join(
                    filter(
                        None,
                        experiment_metadata.get("series_contact_name", "").split(","),
                    )
                ),
                "source": gse,
            },
        )

    else:
        _LOGGER.info(f"Skipping bedset creation for: '{gse_id}'")

    if stop_pipeline:
        pm.stop_pipeline()

    _LOGGER.info(
        f"Processing of '{gse_id}' completed: "
        f"{project_status.number_of_processed}/{project_status.number_of_samples} processed, "
        f"{project_status.number_of_skipped} skipped, {project_status.number_of_failed} failed"
    )
    return project_status



## !This dict was generated with AI under human supervision
GENOME_FILES = {
    "h_kOcvPobU9it_QR1LjPqeNpM56xjEJQ": "hg38",
    "0OgP0NkIM22lVYT5AMmkbb9knKDhk4I6": "hg38",
    "EiFob05aCWgVU_B_Ae0cypnQut3cxUP1": "hg38",
    "DvAlkUMPq7CRnTYAfXGkQmAOfFqVMZHE": "hg19",
    "XJWKh8nsSqBFfcU0DIHMZohYyCWF-vcA": "hg19",
    "MkXYxV2-83BEcPmzskEVGJsJ7Qkb--gX": "hg38",
    "kgTizBhIBf5BeEMTqBuyzRBi2AlwTRj6": "hg19",
    "7hP7E8o-q6H8qcqMHBEIYdbAK49PoEUZ": "hg38",
    "u1HyLgIlq8M_XvEwy0oGqAvKGHJMGtxH": "hg38",
    "H3RA3Jez6oqMOW87LGuwtSgKQqTgWVxx": "hg19",
    "F9zeFn6M4EN4KGAJush7rZEU3GROoeNP": "hg38",
    "q5Gn6dl5HbkZe6sRs9CZVGwGd1-XwyAq": "hg19",
    "svwHqvgassl0loigdqVIQJdjo7NWDIx3": "hg19",
    "oLfPx0NOBKKXMIngGeQ4YewtU4Ge_wKz": "hg38",
    "RGSkyOkQ4qnSLhjrR_e3AI51Ac-RohAL": "hg38",
    "-qU7PUmse_-pilikFDTJKyrt2_QJvUFy": "hg19",
    "ThZcNYiLuWWL86NdJ8dvvJG15K9mW3Fo": "hg19",
    "NTeQ1GQMt2ocCFkS8Z3_qkvetZjabWSt": "hg38",
    "5ryrHdbJHIgyZuE29h5uzITRL4kinZWG": "hg38",
    "a_WL8OC7sFJfjux5m11M2bKl0dYepA1x": "hg38",
    "6pxZqxG0TYtyVb8yp14ONpaxZ8msQqKr": "hg19",
    "tkRdBlSp5hewK4OpEJC87J9pw-ac0vOa": "hg38",
    "k4mLJvbFzZiw3o6SL8hh63V2u7AjDMrE": "hg19",
    "tmnbiAyj2fke68d_TYjq2g487US8C15r": "hg38",
    "H9er8ocYfIN2TOfyf6zMyeXXm7trXGP7": "hg18",
    "EGlYk1stOsAjmTALWkqKDqtZOSkyA_YH": "hg38",
    "lXa-sGAmSafYXHN4iEwup0EsQh5F1krA": "hg19",
    "FnbS0xDAGOePD7lp6Xt0vnbcASzkk3gk": "hg38",
    "vRjC5qM1Tc-fFjJo0TGRw4CVjFhmLG0f": "hg38",
    "wdpZbFN0pd92H2VeKZBAp3riQN4nJXkK": "hg38",
    "GOIHeGSorDrbznRxihs5rIb6vTiTKaw7": "hg19",
    "Dx_M8skbJqROkfXhhQRtWejcCyewRbdL": "hg38",
    "ieWVCws5MC2QFRKgH9QcN3u5_Y_3hPG6": "hg18",
    "jFm0Uca8a7vK2cbuIQgBopjBilgCFheD": "hg38",
    "Ba88PY52_qeifhJrgUXyin6UITdXNsg3": "hg38",
    "YfZ0rklv8KY9DCtqG0iIX16zsgbuBgmM": "hg18",
    "5SdkZCnuZL2YIptqjSBZfupo_O7HpD_B": "hg19",
    "gHcfbUVnFzHv3QSqz2sSqVHdUQbDO8N5": "hg38",
    "eN7J_gZz_meakMCeXXBEvY_njignMPxl": "hg19",
    "6chutju9QVJW0rdA-wgubHbtoTQ42o-6": "hg38",
    "lWRRNMNypacEjnJCy-AYiDNUPy1brQGC": "hg38",
    "xEg2q8K9gV4027DMTCiaUGLCcrqySglR": "hg38",
    "JL56x8L1q1Fs_-jHvZxBG01Vitac-CmO": "hg38",
    "-e70JAQq4NJDg8-1Ab2XhHu6yYjeW-zu": "mm39",
    "FTBYBUoMhkOJ_-8lWpERVTxe62kstAol": "mm10",
    "9k1WrFA4Ys2fPifOOVswhOEdurvsaLfI": "mm10",
    "jpOqOhddb15iOm2SIdJSjsf-U5Uu7Def": "mm38",
    "fXjBOJjw-DYsSnnfDBl5vtZu1N7lbnUl": "mm39",
    "PM8ODmBlTISp4Onv0aSBFaAfi3QVCGzx": "mm39",
    "qcT5VVX5G3mN2O9OqeFR-F0POVuY2oGw": "mm39",
    "99TjKCwZJJLjpBqkLpTgC2E_Y3OgKtMz": "mm10",
    "D-6wf8dsOttiVNnLSImSglRJvw_8Zr_j": "mm39",
    "3MS1-4k87pZ0-C80QDoUvhFmC0usPH28": "mm39",
    "M7ZWnvUTT06JREJnMb_7UGwgGaG0-13s": "mm10",
    "JPyo8AqZzCyVaUx1lAkk6LbpyQPX4VUB": "mm39",
    "qtPKGcXii2OuiyIoDA9K0jSKR62qCyzd": "mm39",
    "2Ls1P5eUdKbvtOhjJx3s2R5r0_I-IB5Z": "mm39",
    "XVhfRj6PCzLoGiEeXjpFv7vriVN02aPc": "mm39",
    "3rgz8-_XPSiTUYPamUTRF3DArhAhTint": "mm39",
    "vygX07e7feibvucSnWj6hRScGMfc7B6P": "mm9",
    "WsOG-InFnIta0rqSy1KUBjrFrukbnE5j": "mm39",
    "KPagVaXI4XwQ1D0L0EW4eEuVl-otaAtX": "mm39",
    "6-UTIAyR94-nanfrhd_sAF6oHLyMd0zH": "mm37",
    "TQORtixTJqM3Su9dmtACKc7hNHAceE4I": "mm38",
    "4mvptys3ckGgiUCcly4HOHB40IhpwwVT": "mm9",
    "dMjpOU7EvpeZVb0gpPoZ7prNaxOu88Ta": "mm10",
    "hW3Ba5zoufl3-MGXQESlWXjsW56R5vPG": "mm10",
    "wsDErYxgCXiPnb2FWZ4sXtx3B0YyruRu": "mm10",
    "bLbjXXCz_5qAonaDXcVuadx65QZkC7mb": "mm39",
    "ABEupc6KHmxtHGarfWFXTmu9mUcNnfoM": "mm39",
}



from typing import Union, Dict


class GenomeModel:
    """
        Initialize genome model

    A class representing a reference genome. You feed it a reference genome. It retrieves chrom sizes (from refgenie),
    and then it provides some helper functions, intended for use with reference genome validation.

    """

    def __init__(
        self,
        genome_alias: str,
        chrom_sizes_file: Union[str, Dict[str, int]],
        genome_digest: Union[str, None] = None,
        # common_aliases: Optional[List] = None,
        # exclude_ranges_names: Optional[List] = None,
    ):
        self._genome_alias = genome_alias
        self.chrom_sizes_file = chrom_sizes_file
        self._genome_digest = genome_digest
        self._chrom_sizes = self.get_chrom_sizes()
        # self.common_aliases = common_aliases  # What are the other names for the other this reference genomes
        # self.excluded_ranges_names = exclude_ranges_names  # Which bed file digests from the excluded ranges are associated with this reference genome?

    @property
    def genome_alias(self):
        return self._genome_alias  # TODO: update everything

    @property
    def genome_digest(self):
        return self._genome_digest

    @property
    def chrom_sizes(self):
        return self._chrom_sizes

    def get_chrom_sizes(self) -> dict:
        """
        Obtains chrom sizes via refgenie (using a refgenconf.refgenconf.RefGenConf object)

        :return dict: dictionary containing chroms(keys) and lengths(values)
        """

        if isinstance(self.chrom_sizes_file, str):
            chrom_sizes_path = self.chrom_sizes_file

            chrom_sizes = {}

            with open(chrom_sizes_path, "r") as f:
                for line in f:
                    chrom, size = line.strip().split("\t")
                    chrom_sizes[chrom] = int(size)
        else:
            # TODO: needs to be validated
            chrom_sizes = self.chrom_sizes_file

        return chrom_sizes

    def filter_excluded_ranges(self, bed_list, igd_hit_matrix):
        """
        BED List of Excluded Ranges files associated with this reference genome.
        These will be manually curated from the Excluded ranges BedSet on BedBase



        """

        # We will probably have a singular .igd database that we will simply compare the bed file to, so this should probably
        # just filter results in a way to determine if there were any hits/not hits for this particular genome

        raise NotImplementedError()



from enum import Enum
from pathlib import Path
from typing import Union

from pydantic import BaseModel, Field, ConfigDict
from gtars.models import RegionSet

from bedboss.models import DATA_FORMAT


class InputTypes(Enum):
    BED_GRAPH = "bedgraph"
    BIG_BED = "bigbed"
    BIG_WIG = "bigwig"
    WIG = "wig"
    BED = "bed"


class BedMakerOutput(BaseModel):
    bed_object: Union[str, RegionSet]
    bed_file: Union[str, Path]
    bigbed_file: Union[str, Path, None] = None
    bed_digest: str = None
    bed_compliance: str = Field(
        default="bed3+0", pattern="^bed(?:[3-9]|1[0-5])(?:\+|$)[0-9]?+$"
    )
    compliant_columns: int
    non_compliant_columns: int
    data_format: DATA_FORMAT

    model_config = ConfigDict(arbitrary_types_allowed=True)



## FILE with functions to download chrom sizes from refgenie and validate them against the genome model

import os
import requests
from pydantic import BaseModel
import json
from typing import List, Union, Any
from tqdm import tqdm
import logging
import warnings

from bedboss.refgenome_validator.genome_model import GenomeModel
from geniml.bbclient.const import DEFAULT_CACHE_FOLDER
from bbconf import BedBaseAgent
from bedboss.const import PKG_NAME
from bedboss.exceptions import BedBossException

BASE_URL = "https://api.refgenie.org"
GENOMES_URL = os.path.join(BASE_URL, "v4/genomes?limit=1000")
SEQ_COL_URL = os.path.join(
    BASE_URL, "seqcol/collection/{digest}?collated=true&attribute=name_length_pairs"
)

_LOGGER = logging.getLogger(PKG_NAME)


class SeqCol(BaseModel):
    length: int
    name: str


class SeqColGenome(BaseModel):
    genome: Union[str, None] = None
    digest: str
    description: str
    collection: List[SeqCol]


class Genomes(BaseModel):
    genomes: List[SeqColGenome]


def run_requests(url, timeout=60) -> Any:
    """
    Run a GET request to the specified URL and return the response.
    """
    try:
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()  # Raise an error for bad responses
        return response.json()
    except requests.RequestException as e:
        _LOGGER.error(f"Error fetching data from {url}: {e}")

        warnings.warn(f"Error fetching data from {url}: {e}")
        return None


def get_genome_list() -> List[dict]:
    """
    Fetch the list of genomes from Refgenie.
    """

    genome_data = run_requests(GENOMES_URL)
    if genome_data:
        return genome_data.get("items", [])
    warnings.warn("Failed to fetch genomes from Refgenie, returning empty list.")
    return []


def seq_col_from_digest(digest: str) -> List[SeqCol]:
    """

    Fetch sequence collection from Refgenie using the genome digest.

    :param digest: The digest of the genome to fetch the sequence collection for.

    :return: A list of SeqCol objects containing sequence names and lengths.
    """
    url = SEQ_COL_URL.format(digest=digest)

    # Getting first level of genome info
    response = run_requests(url)
    return_list = []

    for item in response:
        return_list.append(SeqCol(**item))

    return return_list


def get_seq_col() -> Genomes:
    """
    Fetch sequence collections for all genomes from Refgenie and return them as a Genomes object
    containing a list of SeqColGenome objects.

    This function retrieves the list of genomes, fetches their sequence collections,

    :return: Genomes object containing SeqColGenome objects for each genome.

    """

    genomes = get_genome_list()
    if not genomes:
        raise BedBossException("Failed to fetch genome list from Refgenie.")

    return_list = []
    # genomes = genomes[500:-1]  # TEMP for testing
    for info in tqdm(genomes, desc="Downloading genomes from refgenie..."):
        digest = info["digest"]
        seq_col = seq_col_from_digest(digest)

        return_list.append(
            SeqColGenome(
                genome=info["aliases"][0],
                digest=digest,
                description=info["description"],
                collection=seq_col,
            )
        )

    return Genomes(genomes=return_list)


def save_seq_col_to_json(genomes: Genomes, output_path: str = "genome_seqcol.json"):
    """
    Save sequence collections to a JSON file.

    Args:
        genomes: Genomes object to save
        output_path: Path to save the JSON file
    """

    _LOGGER.info("Saving Refgenie genomes for later reuse...")
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    with open(output_path, "w") as f:
        json.dump(genomes.model_dump(), f, indent=4)
    _LOGGER.info(f"Saved sequence collections to {output_path}")


def read_seq_col_from_json(input_path: str = "genome_seqcol.json") -> Genomes:
    """
    Read sequence collections from a JSON file.

    Args:
        input_path: Path to the JSON file

    Returns:
        Genomes object containing the sequence collections
    """

    # os.makedirs(os.path.dirname(input_path), exist_ok=True)
    with open(input_path, "r") as f:
        data = json.load(f)
    return Genomes(**data)


def modify_for_analysis(genomes: Genomes) -> List[GenomeModel]:
    """
    Modify the genomes data for analysis.

    Args:
        genomes: Genomes object to modify

    Returns:
        List of dictionaries with genome information
    """
    genome_list = []

    for genome in genomes.genomes:
        g_size = {}
        for seq in genome.collection:
            g_size[seq.name] = seq.length

        genome_list.append(
            GenomeModel(
                genome_alias=genome.genome,
                chrom_sizes_file=g_size,
                genome_digest=genome.digest,
            )
        )

    return genome_list


def get_chrom_sizes() -> list[GenomeModel]:
    """
    Get chromosome sizes from Refgenie and return them as a list of GenomeModel objects.

    If the cached file exists, it reads from there; otherwise, it downloads the data and saves it.
    """
    cached_file_path = os.path.join(DEFAULT_CACHE_FOLDER, "genome_seqcol.json")

    try:
        ret = read_seq_col_from_json(input_path=cached_file_path)
    except FileNotFoundError:
        _LOGGER.info("No genome_seqcol.json found, downloading from refgenie...")
        ret = get_seq_col()

        save_seq_col_to_json(ret, output_path=cached_file_path)
    return modify_for_analysis(ret)


def update_db_genomes(bbagent: BedBaseAgent) -> None:
    """
    Update the database with the latest genome information from Refgenie.
    This function fetches the sequence collections and updates the database accordingly.
    """

    _LOGGER.info("Updating database with genome information from Refgenie...")

    from bbconf.db_utils import Session, ReferenceGenome, select

    genome_list = get_chrom_sizes()

    with Session(bbagent.bed._sa_engine) as session:

        available_genomes_statement = select(ReferenceGenome.digest)
        available_genomes_return = session.execute(available_genomes_statement).all()
        available_genomes_list = [item[0] for item in available_genomes_return]

        for genome in genome_list:
            if genome.genome_digest in available_genomes_list:
                continue

            new_genome = ReferenceGenome(
                alias=genome.genome_alias,
                digest=genome.genome_digest,
            )

            session.add(new_genome)
        session.commit()

    _LOGGER.info("Database update completed.")



import logging

from bbconf.bbagent import BedBaseAgent

_LOGGER = logging.getLogger("bedboss")


def add_to_qdrant(config: str, batch: int = 100, purge: bool = False) -> None:
    """
    Reindex all bed files in qdrant

    :param config: path to the config file
    """
    agent = BedBaseAgent(config=config)
    agent.bed.reindex_qdrant(batch=batch, purge=purge)


def reindex_semantic_search(
    config: str, purge: bool = False, batch: int = 1000
) -> None:
    """
    Reindex semantic search in qdrant

    :param config: path to the config file.
    :param purge: whether to purge the existing index before reindexing.
    :param batch: number of items that will be uploaded to qdrant in one batch.

    :return: None
    """

    agent = BedBaseAgent(config=config)
    agent.bed.reindex_hybrid_search(batch=batch, purge=purge)
    _LOGGER.info("Semantic search reindexing completed.")



import logging
import subprocess
from typing import List, Union, Dict, Any

from bedboss.refgenome_validator.models import CompatibilityConcise
from bedboss.refgenome_validator.const import GENOME_FILES
from gtars.models import RegionSet as GRegionSet

_LOGGER = logging.getLogger("bedboss")


def get_bed_chrom_info(bedfile: Union[str, GRegionSet]) -> Dict[str, int]:
    """
    Determine chrom lengths for bed file

    :param bedfile: RegionSet object or path to bed file
    returns dict: returns dictionary where keys are chrom names and values are the max end position of that chromosome.
    """
    # if isinstance(bedfile, RegionSet):
    #     df = bedfile.to_pandas()
    #     max_end_for_each_chrom = df.groupby(0)[2].max().to_dict()
    # elif isinstance(bedfile, GRegionSet):
    #     max_end_for_each_chrom = {}
    #     for region in bedfile:
    #         if region.chrom not in max_end_for_each_chrom:
    #             max_end_for_each_chrom[region.chrom] = region.end
    #         if region.end > max_end_for_each_chrom[region.chrom]:
    #             max_end_for_each_chrom[region.chrom] = region.end
    # else:
    #     df = RegionSet(bedfile).to_pandas()
    #     max_end_for_each_chrom = df.groupby(0)[2].max().to_dict()
    #
    # return max_end_for_each_chrom

    if isinstance(bedfile, GRegionSet):
        return_dict = bedfile.get_max_end_per_chr()
    else:
        rs = GRegionSet(bedfile)
        return_dict = rs.get_max_end_per_chr()
    return return_dict


def run_igd_command(command):
    """
    Run IGD via a subprocess, this is a temp implementation until Rust IGD python bindings are finished.
    """

    result = subprocess.run(command, shell=True, capture_output=True, text=True)
    if result.returncode == 0:
        return result.stdout
    else:
        _LOGGER.error(f"Error running command: {result.stderr}")
        return None


def parse_IGD_output(output_str) -> Union[None, List[dict]]:
    """
    Parses IGD terminal output into a list of dicts
    Args:
      output_str: The output string from IGD

    Returns:
      A list of dictionaries, where each dictionary represents a record.
    """

    try:
        lines = output_str.splitlines()
        data = []
        for line in lines:
            if line.startswith("index"):
                continue  # Skip the header line
            elif line.startswith("Total"):
                break  # Stop parsing after the "Total" line
            else:
                fields = line.split()
                record = {
                    "index": int(fields[0]),
                    "number_of_regions": int(fields[1]),
                    "number_of_hits": int(fields[2]),
                    "file_name": fields[3],
                }
                data.append(record)
        return data
    except Exception:
        return None


def predict_from_compatibility_resutlts(
    compatibility_stats: dict[str, CompatibilityConcise],
) -> tuple[str | Any, str]:
    tier1_genomes: List[tuple[str, CompatibilityConcise]] = [
        (genome, prediction)
        for genome, prediction in compatibility_stats.items()
        if prediction.tier_ranking == 1
    ]

    if not tier1_genomes:
        # Fall back to tier 2 if there's exactly one tier 2 genome and no tier 1 genomes
        tier2_genomes: List[tuple[str, CompatibilityConcise]] = [
            (genome, prediction)
            for genome, prediction in compatibility_stats.items()
            if prediction.tier_ranking == 2
        ]
        if len(tier2_genomes) == 1:
            _LOGGER.info(
                "No tier 1 genomes found, using single tier 2 genome as fallback"
            )
            best_digest = tier2_genomes[0][0]
            genome_id = GENOME_FILES.get(best_digest)
            if genome_id is None:
                return None, None
            return genome_id, best_digest
        return None, None

    tier1_genomes.sort(
        key=lambda x: (
            x[1].xs,
            x[1].oobr if x[1].oobr is not None else 0.0,
            x[1].sequence_fit if x[1].sequence_fit is not None else 0.0,
        ),
        reverse=True,
    )

    best_digest = tier1_genomes[0][0]
    genome_id = GENOME_FILES.get(best_digest)
    if genome_id is None:
        return None, None
    return genome_id, best_digest



import subprocess
import socket
import signal
import os
import re
import time
from typing import Union

from bedboss.const import PKG_NAME
from logging import getLogger

_LOGGER = getLogger(PKG_NAME)

script_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "tools/r-service.R"
)


class RServiceManager:
    """
    A class to manage the lifecycle of an R service, allowing files to be processed through the service.

    Attributes:
        r_script_path (str): Path to the R script that starts the service.
        host (str): Host address for the socket connection.
        port (int): Port number for the socket connection.
        process (subprocess.Popen): The process running the R service.
    """

    def __init__(self, host="127.0.0.1", port=8888):
        """
        Initializes the RServiceManager with the given R script path, host, and port.

        Args:
            r_script_path (str): Path to the R script that starts the service.
            host (str): Host address for the socket connection. Default is "127.0.0.1".
            port (int): Port number for the socket connection. Default is 8888.
        """
        self.r_script_path = script_path
        self.host = host
        self.port = port
        self.process = None

        self.start_service()

    def start_service(self):
        """
        Starts the R service by running the R script in a subprocess.
        """
        _LOGGER.info("RService: Starting R service...")
        cmd = ["Rscript", self.r_script_path]
        self.process = subprocess.Popen(cmd, shell=False, preexec_fn=os.setsid)

        while True:
            if self.check_status() == "idle":
                _LOGGER.info(
                    f"RService: Running R process with PID: {self.process.pid}"
                )
                break
            time.sleep(2)

    def run_file(
        self,
        file_path: str,
        digest: str,
        outpath: str,
        genome: str,
        openSignalMatrix: Union[str, None],
        gtffile: Union[str, None],
    ):
        """
        Sends a file path to the R service for processing.

        :param file_path: Path to the file to be processed.
        :param digest: Digest of the file.
        :param outpath: Path to the output directory.
        :param genome: Genome assembly.
        :param openSignalMatrix: Path to the Open Signal Matrix file.
        :param gtffile: Path to the GTF file.

        :return: None
        :exit: 1 if the connection is refused.
        """
        return self.run_command(
            f"{file_path}, {digest}, {outpath}, {genome}, {openSignalMatrix}, {gtffile}\n"
        )

    def run_command(self, command):
        """
        Sends a command to the R service for processing.
        """
        try:
            s = socket.socket()
            s.connect((self.host, self.port))

            _LOGGER.info(f"RService: Sending command: {command}")
            s.send(command.encode())

            while True:
                msg = self.check_status()
                if msg == "idle":
                    _LOGGER.debug(f"RService: Message recieved: {msg}")
                    break
                else:
                    _LOGGER.info(f"RService: Message recieved: {msg}")
                time.sleep(1)
            s.close()
            return s
        except ConnectionRefusedError:
            # BedBossException("Connection refused. Make sure the R service is running.")
            _LOGGER.error(
                "RService: Connection refused. Make sure the R service is running. Unable to send command."
            )
            exit(1)

    def check_status(self):
        """
        Checks the status of the R service by sending a "check" message to the service.
        """
        try:
            s = socket.socket()
            s.connect((self.host, self.port))
            s.send("check\n".encode())
            msg = s.recv(1024).decode()
            key = re.split(r"[\r\n]", msg)[0]

            _LOGGER.debug(f"RService: Received message: {key}")
            s.close()

            return key
        except ConnectionRefusedError:
            _LOGGER.warning("Connection refused. Make sure the R service is running.")

    def terminate_service(self):
        """
        Terminates the R service by sending a termination signal and ensuring the process is stopped.
        """
        self.run_command("done")  # send secrete "terminate" code
        if self.process:
            self.process.terminate()
            try:
                self.process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
            _LOGGER.info("RService: R process terminated.")

    def __del__(self):
        self.terminate_service()



import logging
from typing import Optional, Union

import pandas as pd

from bedboss.exceptions import BedTypeException
from bedboss.models import BedClassificationOutput, DATA_FORMAT

_LOGGER = logging.getLogger("bedboss")


def get_bed_classification(
    bed: Union[str, pd.DataFrame],
    no_fail: Optional[bool] = True,
) -> BedClassificationOutput:
    """
    Get the BED file classification as a Pydantic object.

    :param bed: path to the bed file OR a dataframe
    :param no_fail: should the function (and pipeline) continue if this function fails to parse BED file
    :return BedClassificationOutput object
    """
    #    column format for bed12
    #    string chrom;       "Reference sequence chromosome or scaffold"
    #    uint   chromStart;  "Start position in chromosome"
    #    uint   chromEnd;    "End position in chromosome"
    #    string name;        "Name of item."
    #    uint score;          "Score (0-1000)"
    #    char[1] strand;     "+ or - for strand"
    #    uint thickStart;   "Start of where display should be thick (start codon)"
    #    uint thickEnd;     "End of where display should be thick (stop codon)"
    #    uint reserved;     "Used as itemRgb as of 2004-11-22"
    #    int blockCount;    "Number of blocks"
    #    int[blockCount] blockSizes; "Comma separated list of block sizes"
    #    int[blockCount] chromStarts; "Start positions relative to chromStart"

    def _read_bed_file(filepath: str, skiprows: int = 0) -> Optional[pd.DataFrame]:
        """
        Helper function to read BED file with error handling.
        :param str file_path: path to the bed file
        :param int skip_rows: how many rows to skip during reading
        :return pd.DataFrame
        """
        try:
            df = pd.read_csv(
                filepath, sep="\t", header=None, low_memory=False, skiprows=skiprows
            )
            if skiprows > 0:
                _LOGGER.info(f"Skipped {skiprows} rows to parse bed file {filepath}")
            return df
        except UnicodeDecodeError:
            try:
                df = pd.read_csv(
                    filepath,
                    sep="\t",
                    header=None,
                    nrows=4,
                    skiprows=skiprows,
                    encoding="utf-16",
                )
                if skiprows > 0:
                    _LOGGER.info(
                        f"Skipped {skiprows} rows to parse bed file {filepath}"
                    )
                return df
            except (pd.errors.ParserError, pd.errors.EmptyDataError):
                return None
        except (pd.errors.ParserError, pd.errors.EmptyDataError):
            return None

    if isinstance(bed, str):
        max_rows = 5
        for row_count in range(max_rows + 1):
            df = _read_bed_file(bed, row_count)
            if df is not None:
                break
        else:
            if no_fail:
                _LOGGER.warning(
                    f"Unable to parse bed file {bed}, setting data_format = unknown_data_format"
                )
                return BedClassificationOutput(
                    bed_compliance="unknown_bed_compliance",
                    data_format="unknown_data_format",
                    compliant_columns=0,
                    non_compliant_columns=0,
                )
            else:
                raise BedTypeException(
                    reason=f"Data format could not be determined for {bed}"
                )
    elif isinstance(bed, pd.DataFrame):
        df = bed
    else:
        if no_fail:
            return BedClassificationOutput(
                bed_compliance="unknown_bed_compliance",
                data_format="unknown_data_format",
                compliant_columns=0,
                non_compliant_columns=0,
            )
        else:
            raise BedTypeException(reason="Input is not a string or dataframe.")

    df = df.dropna(axis=1)
    num_cols = len(df.columns)
    compliant_columns = 0
    bed_format_named = DATA_FORMAT.UCSC_BED
    relaxed = False

    def _check_column(col_index: int, checks: list) -> bool:
        """
        Helper function to perform column checks.

        :param col_index: index of the column
        :param checks: list of check functions

        :return: True if all checks pass, False otherwise

        """
        for check in checks:
            if not check(df[col_index]):
                return False
        return True

    # regex patterns for 255,255,255 or 0 for colors: ([0, 255], [0, 255], [0, 255]) | 0
    REGEX_COLORS = r"^(?:\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])(?:,(?:\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])){0,2}$"

    column_checks = {
        0: [lambda col: col.astype(str).str.match(r"[A-Za-z0-9_]{1,255}").all()],
        1: [lambda col: col.dtype == "int" and (col >= 0).all()],
        2: [lambda col: col.dtype == "int" and (col >= 0).all()],
        3: [lambda col: col.astype(str).str.match(r"[\x20-\x7e]{1,255}").all()],
        4: [
            lambda col: col.dtype == "int" and col.between(0, 1000).all(),
        ],
        5: [lambda col: col.isin(["+", "-", "."]).all()],
        6: [lambda col: col.dtype == "int" and (col >= 0).all()],
        7: [lambda col: col.dtype == "int" and (col >= 0).all()],
        8: [lambda col: col.astype(str).str.match(REGEX_COLORS).all()],
        9: [lambda col: col.dtype == "int"],
        10: [
            lambda col: col.astype(str).str.match(r"^(0(,\d+)*|\d+(,\d+)*)?,?$").all()
        ],
        11: [
            lambda col: col.astype(str).str.match(r"^(0(,\d+)*|\d+(,\d+)*)?,?$").all()
        ],
        12: [lambda col: pd.api.types.is_float_dtype(col.dtype) or (col == -1).all()],
        13: [lambda col: col.dtype == "int" and col.iloc[0] != -1],
    }

    for col_index in range(num_cols):
        checks = column_checks.get(col_index, [])
        if _check_column(col_index, checks) and col_index < 12:
            compliant_columns += 1

        elif (
            col_index == 4 and df[col_index].dtype == "int" and df[col_index].all() >= 0
        ):

            compliant_columns += 1
            relaxed = True

        else:
            nccols = num_cols - compliant_columns
            if col_index >= 6:
                if (
                    num_cols == 10
                    and col_index == 6
                    and _check_column(6, column_checks[12])
                    and _check_column(7, column_checks[12])
                    and _check_column(8, column_checks[12])
                    and _check_column(9, column_checks[9])
                ):
                    bed_format_named = (
                        DATA_FORMAT.ENCODE_NARROWPEAK_RS
                        if relaxed
                        else DATA_FORMAT.ENCODE_NARROWPEAK
                    )
                    return BedClassificationOutput(
                        bed_compliance=f"bed{compliant_columns}+{nccols}",
                        data_format=bed_format_named,
                        compliant_columns=compliant_columns,
                        non_compliant_columns=nccols,
                    )
                elif num_cols == 9 and col_index == 6:
                    if (
                        _check_column(6, column_checks[12])
                        and _check_column(7, column_checks[12])
                        and _check_column(8, column_checks[12])
                    ):
                        bed_format_named = (
                            DATA_FORMAT.ENCODE_BROADPEAK_RS
                            if relaxed
                            else DATA_FORMAT.ENCODE_BROADPEAK
                        )
                        return BedClassificationOutput(
                            bed_compliance=f"bed{compliant_columns}+{nccols}",
                            data_format=bed_format_named,
                            compliant_columns=compliant_columns,
                            non_compliant_columns=nccols,
                        )
                    elif (
                        _check_column(6, column_checks[12])
                        and _check_column(7, column_checks[12])
                        and _check_column(8, column_checks[13])
                    ):
                        bed_format_named = (
                            DATA_FORMAT.ENCODE_RNA_ELEMENTS_RS
                            if relaxed
                            else DATA_FORMAT.ENCODE_RNA_ELEMENTS
                        )
                        return BedClassificationOutput(
                            bed_compliance=f"bed{compliant_columns}+{nccols}",
                            data_format=bed_format_named,
                            compliant_columns=compliant_columns,
                            non_compliant_columns=nccols,
                        )
                elif (
                    num_cols == 15
                    and col_index == 12
                    and _check_column(12, column_checks[12])
                    and _check_column(13, column_checks[12])
                    and _check_column(14, column_checks[12])
                ):
                    bed_format_named = (
                        DATA_FORMAT.ENCODE_GAPPEDPEAK_RS
                        if relaxed
                        else DATA_FORMAT.ENCODE_GAPPEDPEAK
                    )
                    return BedClassificationOutput(
                        bed_compliance=f"bed{compliant_columns}+{nccols}",
                        data_format=bed_format_named,
                        compliant_columns=compliant_columns,
                        non_compliant_columns=nccols,
                    )
            bed_format_named = (
                DATA_FORMAT.BED_LIKE_RS if relaxed else DATA_FORMAT.BED_LIKE
            )
            if relaxed and nccols == 0:
                bed_format_named = DATA_FORMAT.UCSC_BED_RS
            return BedClassificationOutput(
                bed_compliance=f"bed{compliant_columns}+{nccols}",
                data_format=bed_format_named,
                compliant_columns=compliant_columns,
                non_compliant_columns=nccols,
            )

    bed_format_named = DATA_FORMAT.UCSC_BED_RS if relaxed else bed_format_named
    return BedClassificationOutput(
        bed_compliance=f"bed{compliant_columns}+0",
        data_format=bed_format_named,
        compliant_columns=compliant_columns,
        non_compliant_columns=0,
    )



import os

import pytest

from bedboss.bedclassifier import (
    get_bed_classification,
)
from bedboss.models import BedClassificationOutput, DATA_FORMAT

FILE_DIR = os.path.dirname(os.path.realpath(__file__))
HG19_CORRECT_DIR = os.path.join(FILE_DIR, "test_data", "bed", "hg19", "correct")
FILE_PATH = f"{HG19_CORRECT_DIR}/sample1.bed.gz"
FILE_PATH_UNZIPPED = f"{HG19_CORRECT_DIR}/hg19_example1.bed"

SIMPLE_EXAMPLES_DIR = os.path.join(FILE_DIR, "data", "bed", "simpleexamples")
BED1 = f"{SIMPLE_EXAMPLES_DIR}/bed1.bed"
BED2 = f"{SIMPLE_EXAMPLES_DIR}/bed2.bed"
BED3 = f"{SIMPLE_EXAMPLES_DIR}/bed3.bed"
BED_4_PLUS_5 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_4plus5.bed"
BED_4_PLUS_6 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_4plus6.bed"
BED_6_PLUS_4 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_6plus4.bed"
BED_7_PLUS_3 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_7plus3.bed"
BED_10_PLUS_0 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_10plus0.bed"
BED_12_PLUS_0 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_12plus0.bed"
BED_12_PLUS_3 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_12plus3.bed"
BED_NARROWPEAK = f"{SIMPLE_EXAMPLES_DIR}/test_nrwpk.bed"
BED_NONSTRICT_NARROWPEAK = f"{SIMPLE_EXAMPLES_DIR}/test_ns_nrwpk.bed"  # this has values greater than 1000 in (col 5)
BED_RNA_ELEMENTS = f"{SIMPLE_EXAMPLES_DIR}/test_rna_elements.bed"
BED_BROADPEAK = f"{SIMPLE_EXAMPLES_DIR}/test_brdpk.bed"
BED_7_01 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_7_01.bed"
BED_7_02 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_7_02.bed"
BED_7_03 = f"{SIMPLE_EXAMPLES_DIR}/test_bed_7_03.bed"
BED_GAPPED_PEAK = f"{SIMPLE_EXAMPLES_DIR}/example.gappedPeak.gz"
BED_GAPPED_PEAK_RS = f"{SIMPLE_EXAMPLES_DIR}/example_2.gappedPeak.gz"


class TestBedClassifier:
    def test_classification(
        self,
    ):
        bedclass = get_bed_classification(bed=FILE_PATH)

    def test_get_bed_classification(
        self,
    ):
        bedtype = get_bed_classification(bed=FILE_PATH_UNZIPPED)
        assert bedtype == BedClassificationOutput(
            bed_compliance="bed6+3",
            data_format=DATA_FORMAT.ENCODE_BROADPEAK,
            compliant_columns=6,
            non_compliant_columns=3,
        )

    @pytest.mark.parametrize(
        "values",
        [
            (
                BED1,
                BedClassificationOutput(
                    bed_compliance="bed6+4",
                    data_format=DATA_FORMAT.ENCODE_NARROWPEAK,
                    compliant_columns=6,
                    non_compliant_columns=4,
                ),
            ),
            (
                BED2,
                BedClassificationOutput(
                    bed_compliance="bed6+3",
                    data_format=DATA_FORMAT.ENCODE_BROADPEAK,
                    compliant_columns=6,
                    non_compliant_columns=3,
                ),
            ),
            (
                BED3,
                BedClassificationOutput(
                    bed_compliance="bed6+2",
                    data_format=DATA_FORMAT.BED_LIKE,
                    compliant_columns=6,
                    non_compliant_columns=2,
                ),
            ),
            (
                BED_4_PLUS_5,
                BedClassificationOutput(
                    bed_compliance="bed6+3",
                    data_format=DATA_FORMAT.ENCODE_BROADPEAK_RS,
                    compliant_columns=6,
                    non_compliant_columns=3,
                ),
            ),
            (
                BED_4_PLUS_6,
                BedClassificationOutput(
                    bed_compliance="bed6+4",
                    data_format=DATA_FORMAT.BED_LIKE_RS,
                    compliant_columns=6,
                    non_compliant_columns=4,
                ),
            ),
            (
                BED_6_PLUS_4,
                BedClassificationOutput(
                    bed_compliance="bed6+4",
                    data_format=DATA_FORMAT.BED_LIKE,
                    compliant_columns=6,
                    non_compliant_columns=4,
                ),
            ),
            (
                BED_7_PLUS_3,
                BedClassificationOutput(
                    bed_compliance="bed7+3",
                    data_format=DATA_FORMAT.BED_LIKE,
                    compliant_columns=7,
                    non_compliant_columns=3,
                ),
            ),
            (
                BED_7_01,
                BedClassificationOutput(
                    bed_compliance="bed7+0",
                    data_format=DATA_FORMAT.UCSC_BED_RS,
                    compliant_columns=7,
                    non_compliant_columns=0,
                ),
            ),
            (
                BED_7_02,
                BedClassificationOutput(
                    bed_compliance="bed7+0",
                    data_format=DATA_FORMAT.UCSC_BED,
                    compliant_columns=7,
                    non_compliant_columns=0,
                ),
            ),
            (
                BED_7_03,
                BedClassificationOutput(
                    bed_compliance="bed6+1",
                    data_format=DATA_FORMAT.BED_LIKE_RS,
                    compliant_columns=6,
                    non_compliant_columns=1,
                ),
            ),
            (
                BED_10_PLUS_0,
                BedClassificationOutput(
                    bed_compliance="bed10+0",
                    data_format=DATA_FORMAT.UCSC_BED,
                    compliant_columns=10,
                    non_compliant_columns=0,
                ),
            ),
            (
                BED_12_PLUS_0,
                BedClassificationOutput(
                    bed_compliance="bed12+0",
                    data_format=DATA_FORMAT.UCSC_BED,
                    compliant_columns=12,
                    non_compliant_columns=0,
                ),
            ),
            (
                BED_12_PLUS_3,
                BedClassificationOutput(
                    bed_compliance="bed12+3",
                    data_format=DATA_FORMAT.BED_LIKE,
                    compliant_columns=12,
                    non_compliant_columns=3,
                ),
            ),
            (
                BED_NARROWPEAK,
                BedClassificationOutput(
                    bed_compliance="bed6+4",
                    data_format=DATA_FORMAT.ENCODE_NARROWPEAK,
                    compliant_columns=6,
                    non_compliant_columns=4,
                ),
            ),
            (
                BED_NONSTRICT_NARROWPEAK,
                BedClassificationOutput(
                    bed_compliance="bed6+4",
                    data_format=DATA_FORMAT.ENCODE_NARROWPEAK_RS,
                    compliant_columns=6,
                    non_compliant_columns=4,
                ),
            ),
            (
                BED_RNA_ELEMENTS,
                BedClassificationOutput(
                    bed_compliance="bed6+3",
                    data_format=DATA_FORMAT.ENCODE_RNA_ELEMENTS,
                    compliant_columns=6,
                    non_compliant_columns=3,
                ),
            ),
            (
                BED_BROADPEAK,
                BedClassificationOutput(
                    bed_compliance="bed6+3",
                    data_format=DATA_FORMAT.ENCODE_BROADPEAK,
                    compliant_columns=6,
                    non_compliant_columns=3,
                ),
            ),
            (
                BED_GAPPED_PEAK,
                BedClassificationOutput(
                    bed_compliance="bed12+3",
                    data_format=DATA_FORMAT.ENCODE_GAPPEDPEAK,
                    compliant_columns=12,
                    non_compliant_columns=3,
                ),
            ),
            (
                BED_GAPPED_PEAK_RS,
                BedClassificationOutput(
                    bed_compliance="bed12+3",
                    data_format=DATA_FORMAT.ENCODE_GAPPEDPEAK_RS,
                    compliant_columns=12,
                    non_compliant_columns=3,
                ),
            ),
        ],
    )
    def test_get_bed_classifications(self, values):
        # bed1 is encode narrowpeak
        # bed2 is encode broadpeak
        # bed 3 is encode bed6+ (6+2)
        # the others are variations on the columns and ensuring they classify correctly

        bedclass = get_bed_classification(bed=values[0])
        assert bedclass == values[1]

    @pytest.mark.skip(reason="Not implemented")
    def test_from_PEPhub_beds(
        self,
    ):
        """"""
        # TODO implement testing from pephub
        pass


#
# def test_manual_dir_beds():
#     """This test is currently just for local manual testing"""
#     local_dir = "/home/drc/Downloads/test_beds_BED_classifier/"
#     # local_dir = "/home/drc/Downloads/individual_beds/"
#     # local_dir = "/home/drc/Downloads/only_narrowpeaks/"
#     output_dir = "/home/drc/Downloads/BED_CLASSIFIER_OUTPUT/"
#     # local_dir = "/home/drc/Downloads/encode_beds/bedfiles/"
#     # output_dir = "/home/drc/Downloads/encode_beds/output/"
#     # local_dir = "/home/drc/Downloads/single_encode_beds/bedfiles/"
#     # output_dir = "/home/drc/Downloads/single_encode_beds/output/"
#
#     for root, dirs, files in os.walk(local_dir):
#         for file in files:
#             print(file)
#             file_path = os.path.join(root, file)
#             print(file_path)
#             bedclass = BedClassifier(
#                 input_file=file_path, output_dir=output_dir, bed_digest=file
#             )
#             print("\nDEBUG BEDCLASS\n")
#             print(bedclass.bed_type)
#             print("+++++++++++++++++++")
#
#
# #
# if __name__ == "__main__":
#     # test_get_bed_type()
#     # test_classification()
#     test_manual_dir_beds()



import os
from typing import Union

import typer

from bedboss import __version__
from bedboss.bbuploader.cli import app_bbuploader
from pephubclient.helpers import MessageHandler as printm

# commented and made new const here, because it speeds up help function,
# from bbconf.const import DEFAULT_LICENSE
DEFAULT_LICENSE = "DUO:0000042"

app = typer.Typer(pretty_exceptions_short=False, pretty_exceptions_show_locals=False)


def create_pm(
    outfolder: str,
    multi: bool = False,
    recover: bool = True,
    dirty: bool = False,
    pipeline_name: str = "bedboss-pipeline",
):
    import pypiper

    pm_out_folder = outfolder
    pm_out_folder = os.path.join(pm_out_folder, "pipeline_manager")

    pm: pypiper.PipelineManager = pypiper.PipelineManager(
        name=pipeline_name,
        outfolder=pm_out_folder,
        version=__version__,
        multi=multi,
        recover=recover,
        dirty=dirty,
    )
    return pm


options_list = ["bigwig", "bedgraph", "bed", "bigbed", "wig"]


def validate_input_options(option: str):
    if option not in options_list:
        raise typer.BadParameter(
            f"Invalid input type option '{option}'. Options are: {', '.join(options_list)}"
        )
    return option


@app.command(help="Run all the bedboss pipeline for a single bed file")
def run_all(
    input_file: str = typer.Option(
        ...,
        help="Path to the input file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    input_type: str = typer.Option(
        ...,
        help=f"Type of the input file. Options are: {', '.join(options_list)}",
        callback=validate_input_options,
        case_sensitive=False,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    license_id: str = typer.Option(
        DEFAULT_LICENSE,
        help="License ID. If not provided for in PEP"
        "for each bed file, this license will be used",
    ),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
    narrowpeak: bool = typer.Option(False, help="Is the input file a narrowpeak file?"),
    check_qc: bool = typer.Option(True, help="Check the quality of the input file?"),
    chrom_sizes: str = typer.Option(None, help="Path to the chrom sizes file"),
    open_signal_matrix: str = typer.Option(
        None, help="Path to the open signal matrix file"
    ),
    ensdb: str = typer.Option(None, help="Path to the EnsDb database file"),
    just_db_commit: bool = typer.Option(False, help="Just commit to the database?"),
    force_overwrite: bool = typer.Option(
        False, help="Force overwrite the output files"
    ),
    update: bool = typer.Option(
        False,
        help="Update the bedbase database with the new record if it exists. This overwrites 'force_overwrite' option",
    ),
    lite: bool = typer.Option(
        False, help="Run the pipeline in lite mode. [Default: False]"
    ),
    upload_qdrant: bool = typer.Option(False, help="Upload to Qdrant"),
    upload_s3: bool = typer.Option(False, help="Upload to S3"),
    upload_pephub: bool = typer.Option(False, help="Upload to PEPHub"),
    # Universes
    universe: bool = typer.Option(False, help="Create a universe"),
    universe_method: str = typer.Option(
        None, help="Method used to create the universe"
    ),
    universe_bedset: str = typer.Option(
        None, help="Bedset used used to create the universe"
    ),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    """

    Run the bedboss pipeline for a single bed file
    """
    from bbconf.bbagent import BedBaseAgent

    from bedboss.bedboss import run_all as run_all_bedboss

    agent = BedBaseAgent(bedbase_config)

    run_all_bedboss(
        input_file=input_file,
        input_type=input_type,
        outfolder=outfolder,
        genome=genome,
        bedbase_config=agent,
        license_id=license_id,
        rfg_config=rfg_config,
        narrowpeak=narrowpeak,
        check_qc=check_qc,
        chrom_sizes=chrom_sizes,
        open_signal_matrix=open_signal_matrix,
        ensdb=ensdb,
        other_metadata=None,
        lite=lite,
        just_db_commit=just_db_commit,
        force_overwrite=force_overwrite,
        update=update,
        upload_qdrant=upload_qdrant,
        upload_s3=upload_s3,
        upload_pephub=upload_pephub,
        universe=universe,
        universe_method=universe_method,
        universe_bedset=universe_bedset,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(help="Run the all bedboss pipeline for a bed files in a PEP")
def run_pep(
    pep: str = typer.Option(..., help="PEP file. Local or remote path"),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    create_bedset: bool = typer.Option(True, help="Create a new bedset"),
    bedset_heavy: bool = typer.Option(
        False, help="Run the heavy version of the bedbuncher pipeline"
    ),
    bedset_id: Union[str, None] = typer.Option(None, help="Bedset ID"),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
    check_qc: bool = typer.Option(True, help="Check the quality of the input file?"),
    ensdb: str = typer.Option(None, help="Path to the EnsDb database file"),
    just_db_commit: bool = typer.Option(False, help="Just commit to the database?"),
    force_overwrite: bool = typer.Option(
        False, help="Force overwrite the output files"
    ),
    update: bool = typer.Option(
        False,
        help="Update the bedbase database with the new record if it exists. This overwrites 'force_overwrite' option",
    ),
    upload_qdrant: bool = typer.Option(True, help="Upload to Qdrant"),
    upload_s3: bool = typer.Option(True, help="Upload to S3"),
    upload_pephub: bool = typer.Option(True, help="Upload to PEPHub"),
    no_fail: bool = typer.Option(False, help="Do not fail on error"),
    license_id: str = typer.Option(DEFAULT_LICENSE, help="License ID"),
    standardize_pep: bool = typer.Option(False, help="Standardize the PEP using bedMS"),
    lite: bool = typer.Option(
        False, help="Run the pipeline in lite mode. [Default: False]"
    ),
    rerun: bool = typer.Option(False, help="Rerun already processed samples"),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    """
    Run the bedboss pipeline for a bed files in a PEP
    """
    from bedboss.bedboss import insert_pep

    pm = create_pm(
        outfolder=outfolder,
        multi=multi,
        recover=recover,
        dirty=dirty,
        pipeline_name=pep.replace("/", "_").replace(":", "_"),
    )

    insert_pep(
        bedbase_config=bedbase_config,
        output_folder=outfolder,
        pep=pep,
        bedset_id=bedset_id,
        rfg_config=rfg_config,
        create_bedset=create_bedset,
        bedset_heavy=bedset_heavy,
        check_qc=check_qc,
        ensdb=ensdb,
        just_db_commit=just_db_commit,
        force_overwrite=force_overwrite,
        update=update,
        license_id=license_id,
        upload_s3=upload_s3,
        upload_pephub=upload_pephub,
        upload_qdrant=upload_qdrant,
        no_fail=no_fail,
        standardize_pep=standardize_pep,
        lite=lite,
        rerun=rerun,
        pm=pm,
    )

    pm.stop_pipeline()


@app.command(
    help="Run unprocessed files or reprocess them. Currently, only hg38, hg19, and mm10 genomes are supported."
)
def reprocess_all(
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    limit: int = typer.Option(100, help="Limit the number of files to reprocess"),
    no_fail: bool = typer.Option(True, help="Do not fail on error"),
):
    from bedboss.bedboss import reprocess_all as reprocess_all_function

    reprocess_all_function(
        bedbase_config=bedbase_config,
        output_folder=outfolder,
        limit=limit,
        no_fail=no_fail,
    )


@app.command(help="Run unprocessed file, or reprocess it [Only 1 file]")
def reprocess_one(
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    identifier: str = typer.Option(..., help="Identifier of the bed file"),
    # PipelineManager
    multi: bool = typer.Option(True, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    from bedboss.bedboss import reprocess_one as reprocess_one_function

    reprocess_one_function(
        bedbase_config=bedbase_config,
        output_folder=outfolder,
        identifier=identifier,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(help="Reprocess a bedset")
def reprocess_bedset(
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    identifier: str = typer.Option(..., help="Bedset ID"),
    no_fail: bool = typer.Option(True, help="Do not fail on error"),
    heavy: bool = typer.Option(False, help="Run the heavy version of the pipeline"),
):
    from bedboss.bedboss import reprocess_bedset as reprocess_bedset_function

    reprocess_bedset_function(
        bedbase_config=bedbase_config,
        output_folder=outfolder,
        identifier=identifier,
        no_fail=no_fail,
        heavy=heavy,
    )


@app.command(help=f"Create a bed files form a [{', '.join(options_list)}] file")
def make_bed(
    input_file: str = typer.Option(
        ...,
        help="Path to the input file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    input_type: str = typer.Option(
        ...,
        help=f"Type of the input file. Options are: {', '.join(options_list)}",
        callback=validate_input_options,
        case_sensitive=False,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
    narrowpeak: bool = typer.Option(False, help="Is the input file a narrowpeak file?"),
    chrom_sizes: str = typer.Option(None, help="Path to the chrom sizes file"),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    """

    Run the bedboss pipeline for a single bed file
    """
    from bedboss.bedmaker.bedmaker import make_bed as mk_bed_func

    mk_bed_func(
        input_file=input_file,
        input_type=input_type,
        output_path=outfolder,
        genome=genome,
        narrowpeak=narrowpeak,
        rfg_config=rfg_config,
        chrom_sizes=chrom_sizes,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(help="Create a bigbed files form a bed file")
def make_bigbed(
    bed_file: str = typer.Option(
        ...,
        help="Path to the input file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
):
    """

    Run the bedboss pipeline for a single bed file
    """
    from bedboss.bedmaker.bedmaker import make_bigbed as mk_bigbed_func

    mk_bigbed_func(
        bed=bed_file,
        output_path=outfolder,
        genome=genome,
        rfg_config=rfg_config,
    )


@app.command(help="Create the statistics for a single bed file.")
def run_stats(
    bed_file: str = typer.Option(
        ...,
        help="Path to the bed file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    ensdb: str = typer.Option(None, help="Path to the EnsDb database file"),
    open_signal_matrix: str = typer.Option(
        None, help="Path to the open signal matrix file"
    ),
    just_db_commit: bool = typer.Option(False, help="Just commit to the database?"),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    from bedboss.bedstat.bedstat import bedstat

    bedstat(
        bedfile=bed_file,
        genome=genome,
        outfolder=outfolder,
        ensdb=ensdb,
        open_signal_matrix=open_signal_matrix,
        just_db_commit=just_db_commit,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(
    help="Reindex the bedbase database and insert all files to the qdrant database."
)
def reindex(
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    purge: bool = typer.Option(False, help="Purge existing index before reindexing"),
    batch: int = typer.Option(1000, help="Number of items to upload in one batch"),
):
    from bedboss.qdrant_index.qdrant_index import add_to_qdrant

    add_to_qdrant(config=bedbase_config, batch=batch, purge=purge)


@app.command(help="Reindex semantic (text) search.")
def reindex_text(
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    purge: bool = typer.Option(False, help="Purge existing index before reindexing"),
    batch: int = typer.Option(1000, help="Number of items to upload in one batch"),
):
    from bedboss.qdrant_index.qdrant_index import reindex_semantic_search

    return reindex_semantic_search(
        config=bedbase_config,
        purge=purge,
        batch=batch,
    )


@app.command(
    help="Create a bedset from a pep file, and insert it to the bedbase database."
)
def make_bedset(
    pep: str = typer.Option(..., help="PEP file. Local or remote path"),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    bedset_name: str = typer.Option(..., help="Name of the bedset"),
    heavy: bool = typer.Option(False, help="Run the heavy version of the pipeline"),
    force_overwrite: bool = typer.Option(
        False, help="Force overwrite the output files"
    ),
    upload_s3: bool = typer.Option(False, help="Upload to S3"),
    upload_pephub: bool = typer.Option(False, help="Upload to PEPHub"),
    no_fail: bool = typer.Option(False, help="Do not fail on error"),
):
    from bedboss.bedbuncher.bedbuncher import run_bedbuncher_form_pep

    run_bedbuncher_form_pep(
        bedbase_config=bedbase_config,
        bedset_pep=pep,
        output_folder=outfolder,
        bedset_name=bedset_name,
        heavy=heavy,
        upload_pephub=upload_pephub,
        upload_s3=upload_s3,
        no_fail=no_fail,
        force_overwrite=force_overwrite,
    )


@app.command(help="Initialize the new, sample configuration file")
def init_config(
    outfolder: str = typer.Option(..., help="Path to the output folder"),
):
    from bedboss.utils import save_example_bedbase_config

    save_example_bedbase_config(outfolder)


@app.command(help="Delete bed from the bedbase database")
def delete_bed(
    sample_id: str = typer.Option(..., help="Sample ID"),
    config: str = typer.Option(..., help="Path to the bedbase config file"),
):
    from bbconf import BedBaseAgent

    bbagent = BedBaseAgent(config)
    bbagent.bed.delete(sample_id)
    print(f"sample {sample_id} deleted from the bedbase database")


def version_callback(value: bool):
    if value:
        typer.echo(f"Bedboss version: {__version__}")
        raise typer.Exit()


@app.command(help="Delete BedSet from the bedbase database")
def delete_bedset(
    identifier: str = typer.Option(..., help="BedSet ID"),
    config: str = typer.Option(..., help="Path to the bedbase config file"),
):
    from bbconf import BedBaseAgent

    bbagent = BedBaseAgent(config)
    bbagent.bedset.delete(identifier)
    print(f"BedSet {identifier} deleted from the bedbase database")


@app.command(help="Tokenize a bedfile")
def tokenize_bed(
    bed_id: str = typer.Option(
        ...,
        help="Path to the bed file",
    ),
    universe_id: str = typer.Option(
        ...,
        help="Universe ID",
    ),
    cache_folder: str = typer.Option(
        None,
        help="Path to the cache folder",
    ),
    add_to_db: bool = typer.Option(
        False,
        help="Add the tokenized bed file to the bedbase database",
    ),
    bedbase_config: str = typer.Option(
        None,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    overwrite: bool = typer.Option(
        False,
        help="Overwrite the existing tokenized bed file",
    ),
):
    from bedboss.tokens.tokens import tokenize_bed_file

    tokenize_bed_file(
        universe=universe_id,
        bed=bed_id,
        cache_folder=cache_folder,
        add_to_db=add_to_db,
        config=bedbase_config,
        overwrite=overwrite,
    )


@app.command(help="Delete tokenized bed file")
def delete_tokenized(
    universe_id: str = typer.Option(
        ...,
        help="Universe ID",
    ),
    bed_id: str = typer.Option(
        ...,
        help="Bed ID",
    ),
    config: str = typer.Option(
        None,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
):
    from bedboss.tokens.tokens import delete_tokenized

    delete_tokenized(
        universe=universe_id,
        bed=bed_id,
        config=config,
    )


@app.command(help="Convert bed file to universe")
def convert_universe(
    bed_id: str = typer.Option(
        ...,
        help="Path to the bed file",
    ),
    config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    method: str = typer.Option(
        None,
        help="Method used to create the universe",
    ),
    bedset: str = typer.Option(
        None,
        help="Bedset used to create the universe",
    ),
):
    from bbconf.bbagent import BedBaseAgent

    bbagent = BedBaseAgent(config)
    bbagent.bed.add_universe(
        bedfile_id=bed_id,
        bedset_id=bedset,
        construct_method=method,
    )


@app.command(help="Update reference genomes in the database")
def update_genomes(
    config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
):
    from bbconf.bbagent import BedBaseAgent
    from bedboss.refgenome_validator.refgenie_chrom_sizes import update_db_genomes

    bbagent = BedBaseAgent(config)
    update_db_genomes(bbagent)

    print("Genomes updated successfully.")


@app.command(help="Download UMAP")
def download_umap(
    config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    output_file: str = typer.Option(
        ...,
        help="Path to the output json file where UMAP embeddings will be saved. (*Python version will be added to the filename)",
    ),
    n_components: int = typer.Option(
        2,
        help="Number of UMAP components",
    ),
    plot_name: str = typer.Option(
        None,
        help="Name of the plot file",
    ),
    plot_label: str = typer.Option(
        None,
        help="Label for the plot",
    ),
    top_assays: int = typer.Option(
        15,
        help="Number of top assays to include",
    ),
    top_cell_lines: int = typer.Option(
        15,
        help="Number of top cell lines to include",
    ),
    method: str = typer.Option(
        "umap",
        help="Dimensionality reduction method to use. Options: 'umap', 'pca', or 'tsne'. To use UMAP, 'umap-learn' package must be installed.",
    ),
):
    from bedboss.scripts.make_umap import get_embeddings

    get_embeddings(
        bbconf=config,
        output_file=output_file,
        n_components=n_components,
        plot_name=plot_name,
        plot_label=plot_label,
        top_assays=top_assays,
        top_cell_lines=top_cell_lines,
        method=method,
    )


@app.command(help="Check installed R packages")
def check_requirements():
    from bedboss.bedboss import requirements_check

    print("Checking pipelines requirements...")
    requirements_check()


@app.command(help="Install R dependencies")
def install_requirements():
    import subprocess

    r_path = os.path.abspath(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "scripts", "installRdeps.R"
        )
    )

    subprocess.run(
        ["Rscript", r_path],
    )


@app.command(help="Verify configuration file")
def verify_config(
    config: str = typer.Argument(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
):
    from bbconf.config_parser.utils import config_analyzer

    try:
        config_analyzer(config)
    except Exception as e:
        printm.print_error(f"Error in provided configuration file: {e}")
        raise typer.Exit(code=1)
    typer.Exit(code=0)


@app.command(help="Get available commands", hidden=True)
def get_commands():
    print(
        " ".join([k.callback.__name__ for k in app.registered_commands]).replace(
            "_", "-"
        )
    )


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", "-v", callback=version_callback, help="App version"
    ),
):
    pass


app.add_typer(app_bbuploader, name="geo")



from typing import Union

from pydantic import BaseModel, ConfigDict


class ChromNameStats(BaseModel):
    xs: float = 0.0
    q_and_m: float = 0.0
    q_and_not_m: float = 0.0
    not_q_and_m: float = 0.0
    jaccard_index: float = 0.0
    jaccard_index_binary: float = 0.0
    passed_chrom_names: bool = False


class ChromLengthStats(BaseModel):
    oobr: Union[float, None] = None
    beyond_range: bool = False
    num_of_chrom_beyond: int = 0
    percentage_bed_chrom_beyond: float = 0.0
    percentage_genome_chrom_beyond: float = 0.0


class SequenceFitStats(BaseModel):
    sequence_fit: Union[float, None] = None


class RatingModel(BaseModel):
    assigned_points: int
    tier_ranking: int
    # model_config = ConfigDict(extra="forbid")


class CompatibilityStats(BaseModel):
    chrom_name_stats: ChromNameStats
    chrom_length_stats: ChromLengthStats
    chrom_sequence_fit_stats: SequenceFitStats
    igd_stats: Union[dict, None] = None
    compatibility: Union[RatingModel, None] = None

    model_config = ConfigDict(extra="forbid")


class CompatibilityConcise(BaseModel):
    xs: float = 0.0
    oobr: Union[float, None] = None
    sequence_fit: Union[float, None] = None
    assigned_points: int
    tier_ranking: int
